from osbot_aws.aws.bedrock.cache.Bedrock__Cache import Bedrock__Cache
from osbot_utils.helpers.html.Tag__Base import Tag__Base
from osbot_utils.helpers.html.Tag__Div import Tag__Div
from osbot_utils.helpers.html.Tag__H import Tag__H
from osbot_utils.helpers.html.Tag__HR import Tag__HR
from osbot_utils.helpers.html.Tag__Head import Tag__Head
from osbot_utils.helpers.html.Tag__Html import Tag__Html


class Bedrock_Cache__Html:

    def __init__(self):
        self.cache        = Bedrock__Cache()
        self.table        = self.cache.cache_table
        self.target_file  = '/tmp/tmp-bedrock-cache.html'
        self.html_tag     = Tag__Html()
        self.title        = 'Html Page with Bedrock_Cache_Data'
        self.sub_title    = '....'
        self.row_elements = []

    def append(self, element):
        self.row_elements.append(element)
        return self

    def create_page(self):
        div_container = Tag__Div(tag_classes=['container-fluid','my-5'])
        h_title       = Tag__H(1, self.title)
        hr            = Tag__HR()
        div_subtitle  = Tag__Div(tag_classes=['badge', 'bg-dark'], inner_html=self.sub_title)
        div_row       = Tag__Div(tag_classes=['row'], elements=self.row_elements)
        div_container.append(h_title,
                             div_subtitle,
                             hr,
                             div_row)

        css_data             = { ".base64-image"  : { "width"         : "200px"           ,
                                                      "height"        : "auto"            ,
                                                      "margin-bottom" : "1rem"            },
                                 ".col"           : { "border"        : "2px solid #C0C0FF",
                                                      "padding"       : "10px"            },
                                 ".bg-dark"       : { "font-size"     : "15px"            },
                                 ".var_name"      : {"font-size"      : "12px"            },
                                 ".comments_value": {"font-size"      : "12px"            }}

        head_style     = Tag__Base(tag_name='style')
        head_tag       = Tag__Head(elements= [head_style])
        head_tag.title = self.title
        head_tag.add_css_bootstrap()
        head_tag.style.set_css(css_data)

        self.html_tag.head  = head_tag
        self.html_tag.body.append(div_container )

    def create_html_file(self):
        self.create_page()
        self.html_tag.save(self.target_file)


from osbot_aws.AWS_Config import AWS_Config
from osbot_aws.apis.shell.Lambda_Shell import Lambda_Shell
from osbot_aws.apis.shell.Shell_Server import Shell_Server
from osbot_aws.apis.Lambda import Lambda                    # todo: see how to resolve the circular dependency with the Lambda Function
from osbot_utils.decorators.methods.cache_on_self import cache_on_self
from osbot_utils.utils.Functions import function_source_code


class Shell_Client:

    def __init__(self, aws_lambda: Lambda):
        self.aws_lambda = aws_lambda

    @cache_on_self
    def _lambda_auth_key(self):
        return Lambda_Shell().get_lambda_shell_auth()

    def _invoke(self, method_name, method_kwargs=None,return_logs=False):
        event = {'lambda_shell': {'method_name': method_name, 'method_kwargs': method_kwargs}}
        if self.aws_lambda:
            event['lambda_shell']['auth_key'] = self._lambda_auth_key()
            if return_logs:
                return self.aws_lambda.invoke_return_logs(event)
            else:
                return self.aws_lambda.invoke(event)
        else:
            return Shell_Server().invoke(event)

    def bash(self,command, cwd=None):
        return self._invoke('bash', {'command': command, 'cwd': cwd})

    def exec(self, executable, params=None, cwd=None):
        result = self.process_run(executable, params, cwd)
        std_out = result.get('stdout', '').strip()
        std_err = result.get('stderr', '').strip()
        std_console = std_out + std_err
        if std_console:
            return std_console
        if result.get('errorMessage'):
            return f'Error: {result.get("errorMessage")}'
        return result

    def exec_function(self, function, return_logs=False):
        return self.python_exec_function(function,return_logs=return_logs)

    def exec_function_return_logs(self, function):
        return self.python_exec_function(function,return_logs=True)

    def process_run(self, executable, params=None, cwd=None):
        return self._invoke('process_run', {'executable' : executable , 'params': params, 'cwd': cwd} )

    def reset(self):
        if self.aws_lambda.s3_bucket is None:                                                                     # if these values are not set
            self.aws_lambda.set_s3_bucket(AWS_Config().lambda_s3_bucket())                                               # use default values
            self.aws_lambda.set_s3_key   (f'{AWS_Config().lambda_s3_folder_lambdas()}/{self.aws_lambda.original_name}.zip')  # which are needed
        return self.aws_lambda.update_lambda_code()                                                               # to trigger the update (which will reset the lambda and force a cold start on next lambda invocation)

    def python_exec(self, code, return_logs=False):
        return self._invoke('python_exec', {'code' : code}, return_logs=return_logs)

    def python_exec_function(self, function, return_logs=False):
        function_name = function.__name__
        function_code = function_source_code(function)
        exec_code     = f"{function_code}\nresult= {function_name}()"
        return self.python_exec(exec_code, return_logs=return_logs)

    # command methods

    def ls(self, path='', cwd='.'):         # todo: fix vuln: this method allows extra process executions via ; and |
        return self.bash(f'ls {path}', cwd).get('stdout')


    def file_contents(self, path):return self._invoke('file_contents',{'path': path})

    # with not params

    def disk_space    (self): return self._invoke('disk_space'    )
    def list_processes(self): return self._invoke('list_processes')
    def memory_usage  (self): return self._invoke('memory_usage'  )
    def ping          (self): return self._invoke('ping'          )
    def ps            (self): return self.exec   ('ps'            )
    def pwd           (self): return self._invoke('pwd'           )
    def whoami        (self): return self.exec   ('whoami'        )


from osbot_aws.AWS_Config import AWS_Config
from osbot_aws.aws.iam.IAM_Role import IAM_Role


class Temp_Aws_Roles:
    def __init__(self):
        self.temp_prefix                      = 'temp_role_'
        self.role_name__for_lambda_invocation = self.temp_prefix + 'for_lambda_invocation' # todo: improve this logic
        pass

    def for_lambda_invocation__create(self, delete_existing=False):
        iam_role = IAM_Role(self.role_name__for_lambda_invocation)
        if delete_existing:
            iam_role.iam.role_delete()
        return iam_role.create_for__lambda().get('role_arn')

    def for_lambda_invocation__not_exists(self):
        return self.for_lambda_invocation_exists() is False

    def for_lambda_invocation_exists(self):
        iam_role = IAM_Role(self.role_name__for_lambda_invocation)
        return iam_role.iam.role_exists()

    def for_lambda_invocation__role_arn(self):
        account_id = AWS_Config().aws_session_account_id()
        role_name  = self.role_name__for_lambda_invocation
        return f"arn:aws:iam::{account_id}:role/{role_name}"



from osbot_aws.aws.bedrock.models.diffusion.Stability_Stable_Diffusion import Stability_Stable_Diffusion
from osbot_aws.aws.bedrock.models.diffusion.Stability_Stable_Diffusion_XL_V1 import Stability_Stable_Diffusion_XL_V1
from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self
from osbot_utils.utils.Png import save_png_base64_to_file


class Stability_Stable_Diffusion_XL_V1__Text_To_Image(Stability_Stable_Diffusion_XL_V1):
    height  : int      = 512
    width   : int      = 512

    def body(self):
        return dict(**super().body()                                   ,
                    height                  = self.height              ,
                    width                   = self.width               )





import  gzip
import  json
import  os
import  tempfile
from    io                                                  import BytesIO
from    gzip                                                import GzipFile
from    boto3.s3.transfer                                   import TransferConfig
from    botocore.errorfactory                               import ClientError

from osbot_aws.aws.boto3.View_Boto3_Rest_Calls import print_boto3_calls
from    osbot_utils.base_classes.Type_Safe                  import Type_Safe
from    osbot_utils.decorators.lists.group_by               import group_by
from    osbot_utils.decorators.lists.index_by               import index_by
from    osbot_aws.apis.Session                              import Session
from    osbot_utils.decorators.methods.cache_on_self        import cache_on_self
from    osbot_utils.decorators.methods.remove_return_value  import remove_return_value
from    osbot_utils.testing.Temp_File                       import Temp_File
from    osbot_utils.utils.Files                             import Files, file_extension
from    osbot_utils.utils.Zip                               import zip_folder

S3_DEFAULT_FILE_CONTENT_TYPE = 'binary/octet-stream'
S3_HTML_PAGE_CONTENT_TYPE    = 'text/html; charset=utf-8'
S3_FILES_CONTENT_TYPES       = { '.js'  : 'application/javascript; charset=utf-8',
                                 '.jpg' : 'image/jpeg'                           ,
                                 '.jpeg': 'image/jpeg'                           ,
                                 '.png' : 'image/png'                            ,
                                 '.txt' : 'text/plain; charset=utf-8'            ,
                                 '.pdf' : 'application/pdf'                      ,
                                 '.html': 'text/html; charset=utf-8'             ,
                                 '.css' : 'text/css; charset=utf-8'              ,
                                 '.svg' : 'image/svg+xml'                        ,
                                 '.gif' : 'image/gif'                            ,
                                 '.webp': 'image/webp'                           ,
                                 '.json': 'application/json; charset=utf-8'      ,
                                 '.xml' : 'application/xml; charset=utf-8'       ,
                                 '.zip' : 'application/zip'                      ,
                                 '.mp3' : 'audio/mpeg'                           ,
                                 '.mp4' : 'video/mp4'                            ,
                                 '.avi' : 'video/x-msvideo'                      ,
                                 '.mov' : 'video/quicktime'                      }


# helper methods

def s3_file_download(s3_bucket, s3_key, use_cache = False):
    return S3().file_download(s3_bucket, s3_key, use_cache)

def s3_file_download_to(s3_bucket, s3_key, target_file, use_cache = False):
    return S3().file_download_to(s3_bucket, s3_key,target_file, use_cache)


class S3(Type_Safe):
    aws_access_key_id       : str = None                                # todo: refactor these client setup details into a common base class since (this is should be usable by all osbot-aws boto3 wrapper classes)
    aws_secret_access_key   : str = None
    endpoint_url            : str = None
    region_name             : str = None
    service_name            : str = 's3'
    tmp_file_folder         : str = 's3_temp_files'
    use_threads             : bool = True


    @cache_on_self
    def client(self):
        kwargs = dict(aws_access_key_id     = self.aws_access_key_id    ,
                      aws_secret_access_key = self.aws_secret_access_key,
                      endpoint_url          = self.endpoint_url         ,
                      region_name           = self.region_name          ,
                      service_name          = self.service_name         )
        return self.session().client(**kwargs)

    # todo: see if we still need this code below to set the region, since the session.client_boto3 should do it
    # @cache_on_self
    # def client(self):
    #     # todo: understand the impact of having to add the endpoint_url below
    #     #       we needed this because (some, not all, which is really weird) buckets
    #     #       were throwing this error when creating the pre-signed URL
    #     #       <Code>SignatureDoesNotMatch</Code>
    #     #           <Message>The request signature we calculated does not match the signature you provided. Check your key and signing method.</Message>
    #     #       this is the comment that gave the solution https://github.com/boto/boto3/issues/1149#issuecomment-793737086
    #     region_name = aws_config.region_name()
    #     # endpoint_url = f'https://s3.{region_name}.amazonaws.com'            # this was causing issues uploading files
    #     # return boto3.client('s3', region_name=region_name,  endpoint_url=endpoint_url)
    #     return boto3.client('s3', region_name=region_name)

    @cache_on_self
    def s3(self):                               # todo replace usages below with this method
        return self.client()

    @cache_on_self
    def s3_resource(self):                                      # todo refactor this method to be called resource()
        return self.session().resource('s3')

    @cache_on_self
    def session(self):
        return Session()

    def bucket_delete_all_files(self, bucket):
        s3_keys = self.find_files(bucket)
        if s3_keys:
            result = self.files_delete(bucket, s3_keys)
            return result
        return True

    @remove_return_value(field_name='ResponseMetadata')
    def bucket_notification(self, bucket_name):
        #if self.boto_notification is None : self.boto_notification = self.s3_resource().BucketNotification(bucket_name=bucket_name)
        #return self.boto_notification
        return self.s3().get_bucket_notification_configuration(Bucket=bucket_name)

    def bucket_notification_set_lambda_permission(self, s3_bucket, lambda_arn):
        from osbot_aws.apis.Lambda import Lambda
        statement_id = 'allow_s3_notifications_to_invoke_function'
        action       = 'lambda:InvokeFunction'
        principal    = 's3.amazonaws.com'
        aws_lambda   = Lambda()
        aws_lambda.permission_delete(lambda_arn, statement_id)
        return aws_lambda.permission_add(lambda_arn, statement_id, action, principal, )

    def bucket_notification_create(self, s3_bucket, notification_config):
        try:

            params = { 'Bucket'                   : s3_bucket,
                       'NotificationConfiguration': notification_config }
            # todo: refactor code bellow to an Lambda specific class and tests
            # self.bucket_notification_set_lambda_permission(s3_bucket, lambda_arn)  # configure permissions
               # {
               #  'LambdaFunctionConfigurations': [{
               #                                    'LambdaFunctionArn': lambda_arn,
               #                                    'Events'           : events,
               #                                    'Filter'           : { 'Key': { 'FilterRules': [ {'Name' : 'prefix',
               #                                                                                     'Value': prefix }]}}}]}}
            return self.s3().put_bucket_notification_configuration(**params)
        except Exception as error:
            return {'error': f'{error}'}

    # main methods
    def bucket_arn(self,bucket):
        return 'arn:aws:s3:::{0}'.format(bucket)

    def bucket_create(self, bucket, region):
        try:
            location=self.s3().create_bucket(Bucket=bucket,CreateBucketConfiguration={'LocationConstraint': region }).get('Location')
            return { 'status':'ok', 'data':location}
        except Exception as error:
            return {'status': 'error', 'data': '{0}'.format(error)}

    def bucket_delete(self, bucket):
        if self.bucket_exists(bucket) is False:
            return False
        self.s3().delete_bucket(Bucket=bucket)
        return self.bucket_exists(bucket) is False

    def bucket_exists(self, bucket_name):
        try:
            self.s3().head_bucket(Bucket=bucket_name)
            return True
        except:
            return False

    def bucket_not_exists(self, bucket_name):
        return self.bucket_exists(bucket_name) is False

    def buckets(self):
        data = []
        buckets = self.s3().list_buckets()
        if buckets:
            for bucket in buckets.get('Buckets'):
                data.append(bucket['Name'])
        return sorted(data)
        #return sorted(list({ bucket['Name'] for bucket in self.s3().list_buckets().get('Buckets')}))

    def create_pre_signed_url(self, bucket_name, object_name, operation='get_object', expiration=3600):     # 3600 is one hour (60 seconds * 60 minutes)
        response = self.client().generate_presigned_url(operation, Params={'Bucket': bucket_name, 'Key': object_name}, ExpiresIn=expiration)
        return response

    def dont_use_threads(self):
        self.use_threads = False
        return self

    def files_raw(self, bucket, prefix='', filter=''):
        params = {'Bucket': bucket}
        if isinstance(prefix, str):                                            # if prefix is a string, filter on list_objects_v2
            params['Prefix'] = prefix

        while True:
            resp = self.s3().list_objects_v2(**params)
            try:
                contents = resp['Contents']
            except KeyError:
                return

            for obj in contents:
                key = obj['Key']
                if (not filter) or (filter and filter in key):
                    yield obj
            try:
                params['ContinuationToken'] = resp['NextContinuationToken']     # token for next batch of 1000 results
            except KeyError:
                break

    def find_files(self, bucket, prefix='', filter=''):
        return list({ item['Key'] for item in self.files_raw(bucket, prefix, filter) })

    def file_bytes(self, bucket, key):
        obj = self.s3().get_object(Bucket=bucket, Key=key)                      # get object data from s3
        return obj['Body'].read()                                               # returns all bytes

    def file_content_type(self, bucket, key):
        return self.file_details(bucket, key).get('ContentType')

    def file_content_type_update(self, bucket, key, metadata, content_type):
        file_copy_kwargs = dict(bucket_source      = bucket      ,
                                key_source         = key         ,
                                bucket_destination = bucket      ,
                                key_destination    = key         ,
                                metadata           = metadata    ,      # in S3 we can't update the content type with also updating the metadata
                                content_type       = content_type)      # i guess this is related to the fact that the content_type is stored as a metadata item
        return self.file_copy(**file_copy_kwargs)


    def file_contents(self, bucket, key, encoding='utf-8'):
        return self.file_bytes(bucket,key).decode(encoding)                      # extract body and decode it

    def file_contents_from_gzip(self, bucket, key):
        obj = self.s3().get_object(Bucket=bucket, Key=key)                    # get object data from s3
        bytestream = BytesIO(obj['Body'].read())
        return GzipFile(None, 'rb', fileobj=bytestream).read().decode('utf-8')

    def file_download(self, bucket, key, use_cache = False):
        if bucket and key:
            tmp_file = '/tmp/' + key.replace('/','_')
            if self.file_download_to(bucket, key, tmp_file, use_cache):
                return tmp_file
        return None

    def file_download_and_delete(self, bucket, key):
        tmp_file = '/tmp/' + key.replace('/','_')
        if self.file_download_to(bucket, key, tmp_file):
            self.file_delete(bucket,key)
            return tmp_file
        return None

    def file_download_to(self, bucket, key, target_file, use_cache = False):
        try:
            if use_cache is True:
                if os.path.exists(target_file):
                    return True
            else:
                if os.path.exists(target_file):
                    os.remove(target_file)
            self.s3().download_file(bucket, key, target_file)
            return os.path.exists(target_file)
        except ClientError:                                     # when file not found
            return False

    @remove_return_value('ResponseMetadata')
    def file_copy(self, bucket_source, key_source, bucket_destination, key_destination, metadata=None,
                  content_type=None):
        kwargs_file_copy = dict(CopySource={'Bucket': bucket_source, 'Key': key_source},
                                Bucket=bucket_destination,
                                Key=key_destination)
        if metadata is not None:
            kwargs_file_copy['Metadata'] = metadata
            kwargs_file_copy['MetadataDirective'] = 'REPLACE'
        if content_type is not None:
            kwargs_file_copy['ContentType'] = content_type
        return self.client().copy_object(**kwargs_file_copy)

    def file_create_from_bytes(self, file_bytes, bucket, key):
        return self.file_upload_from_bytes(file_bytes, bucket, key)

    def file_create_from_string_as_gzip(self, file_contents, bucket, key):
        tmp_path = tempfile.NamedTemporaryFile().name

        with gzip.open(tmp_path,'w') as file:                       # create gzip on tmp_file
            file.write(file_contents.encode())                      # write contents
        self.s3().upload_file(tmp_path, bucket,key)               # upload file using s3 api
        os.remove(tmp_path)                                         # delete tmp file
        return True


    def file_create_from_string (self, file_contents, bucket, key):
        with Temp_File(contents=file_contents, extension=file_extension(key)) as temp_file:
            return self.file_upload_to_key(temp_file.file_path, bucket,key)


    def file_delete(self, bucket, key):
        result = self.client().delete_object(Bucket=bucket, Key=key)              # delete file from s3
        return result.get('ResponseMetadata').get('HTTPStatusCode') == 204

    def files_delete(self, bucket, keys):
        delete_keys = {'Objects': [{'Key': key } for key  in keys]}
        result      = self.client().delete_objects(Bucket=bucket, Delete=delete_keys)     # todo: see if there is a way to get feedback on the delete status
        return result.get('ResponseMetadata').get('HTTPStatusCode') == 200  #       since we also get 200 when the key value is wrong

    @remove_return_value('ResponseMetadata')
    def file_details(self, bucket, key):
        return self.client().head_object(Bucket = bucket, Key= key)

    def file_metadata(self, bucket, key):
        return self.file_details(bucket, key).get('Metadata')

    def file_metadata_update(self, bucket, key, metadata):
        file_copy_kwargs = dict(bucket_source      = bucket  ,
                                key_source         = key     ,
                                bucket_destination = bucket  ,
                                key_destination    = key     ,
                                metadata           = metadata)
        return self.file_copy(**file_copy_kwargs)

    def file_size_in_Mb(self, bucket, key):
        details = self.file_details(bucket,key)
        return '{0:.1f} Mb'.format(int(details['ResponseMetadata']['HTTPHeaders']['content-length']) / 1024 / 1024)

    def file_exists(self, bucket, key):
        try:
            self.s3().head_object(Bucket=bucket, Key=key)                     # calls head_object function
            return True                                                       # if it works, file exists (return true)
        except:                                                               # if we get an exception
            return False                                                      # file doesn't exists (return false)

    def file_move(self,src_bucket, src_key, dest_bucket, dest_key):
        self.file_copy(src_bucket, src_key, dest_bucket, dest_key)
        self.file_delete(src_bucket, src_key)
        return self.file_exists(dest_bucket, dest_key)

    def file_not_exists(self,bucket, key):
        return self.file_exists(bucket=bucket, key=key) is False

    def file_upload(self,file , bucket, folder):
        if not os.path.isfile(file):                                            # check that file to upload exists locally
            return None                                                         # if not return None

        key = os.path.join(folder, os.path.basename(file))                      # create key from folder and file's filename

        self.file_upload_to_key(file, bucket, key)                                # upload file
        return key                                                              # return path to file uploaded (if succeeded)

    def file_upload_from_bytes(self, file_body, bucket, key):
        self.s3().put_object(Body=file_body, Bucket=bucket, Key=key)
        return True

    # def file_upload_to_key(self, file, bucket, key):
    #     self.s3().upload_file(file, bucket, key)                                # upload file
    #     return True                                                             # return true (if succeeded)

    def file_upload_to_key(self, file, bucket, key, set_content_type=True):
        extra_args   = None
        if set_content_type:
            extension = file_extension(file)
            if extension:
                content_type = S3_FILES_CONTENT_TYPES.get(extension)
                if content_type:
                    extra_args = {'ContentType': content_type}
        upload_args   = [file, bucket, key, extra_args]
        upload_kwargs =  dict(Config=self.transfer_config())
        self.client().upload_file(*upload_args, **upload_kwargs)      # upload file
        return True

    def file_upload_as_temp_file(self, file, bucket):
        key = '{0}/{1}'.format(self.tmp_file_folder, Files.temp_filename(Files.file_extension(file)))
        self.file_upload_to_key(file, bucket, key)
        return key

    def folder_contents(self, s3_bucket, parent_folder='', return_full_path=False):
        folders = []
        files = []
        paginator = self.client().get_paginator('list_objects_v2')
        if not parent_folder.endswith('/'):
            parent_folder += '/'

        operation_parameters = {
            'Bucket': s3_bucket,
            'Delimiter': '/'
        }
        if parent_folder != '/':
            operation_parameters['Prefix'] = parent_folder              # add support for also listing files in the root of the bucket

        for page in paginator.paginate(**operation_parameters):
            # Loop through and add files
            for obj in page.get('Contents', []):
                file_path = obj.get('Key')
                if return_full_path:
                    files.append(file_path)
                else:
                    file_name = file_path.replace(parent_folder, '')
                    if file_name:  # Ensure we don't add the parent folder itself
                        files.append(file_name)

            # Loop through and add subfolders
            for common_prefix in page.get('CommonPrefixes', []):
                folder_path = common_prefix.get('Prefix')
                if return_full_path:
                    folders.append(folder_path)
                else:
                    folder_name = folder_path.replace(parent_folder, '').rstrip('/')
                    folders.append(folder_name)

        return {'folders': folders, 'files': files}


    def folder_files(self, s3_bucket, parent_folder='', return_full_path=False):
        files = []
        paginator = self.client().get_paginator('list_objects_v2')
        if not parent_folder.endswith('/'):
            parent_folder += '/'

        operation_parameters = {
            'Bucket': s3_bucket,
            'Delimiter': '/'
        }
        if parent_folder != '/':
            operation_parameters['Prefix'] = parent_folder              # add support for also listing files in the root of the bucket

        for page in paginator.paginate(**operation_parameters):
            # Loop through the files in the specified folder
            for obj in page.get('Contents', []):
                file_path = obj.get('Key')
                if return_full_path:
                    files.append(file_path)
                else:
                    file_name = file_path.replace(parent_folder, '')
                    files.append(file_name)

        return files

    def folder_list(self, s3_bucket, parent_folder='', return_full_path=False):
        folders = []
        paginator = self.client().get_paginator('list_objects_v2')
        if parent_folder.endswith('/') is False:
            parent_folder += '/'

        operation_parameters = {
            'Bucket': s3_bucket    ,
            'Delimiter': '/'
        }
        if parent_folder != '/':
            operation_parameters['Prefix'] = parent_folder              # add support for also listing files in the root of the bucket

        for page in paginator.paginate(**operation_parameters):
            for common_prefix in page.get('CommonPrefixes', []):
                folder_path = common_prefix.get('Prefix')
                if return_full_path:
                    folders.append(folder_path)
                else:
                    folder_name = folder_path.replace(parent_folder, '').replace('/', '')
                    folders.append(folder_name)
        return folders

    def folder_upload (self, folder, s3_bucket, s3_key):
        file = zip_folder(folder)
        self.file_upload_to_key(file, s3_bucket, s3_key)
        os.remove(file)
        return self

    def policy(self, s3_bucket):
        try:
            return json.loads(self.s3().get_bucket_policy(Bucket=s3_bucket).get('Policy'))
        except:
            return {'Statement': []}

    @index_by
    @group_by
    def policy_statements(self, s3_bucket):
        return self.policy(s3_bucket).get('Statement')

    def policy_statements__resource_arn(self, s3_bucket, trail_name, account_id):
        return f'arn:aws:s3:::{s3_bucket}/{trail_name}/AWSLogs/{account_id}/*'

    def policy_statements__new(self, action, effect, service, resource_arn ):
        return { 'Action'   : action,
                 'Effect'   : effect,
                 'Principal': {'Service': service },
                 'Resource' : resource_arn}

    def policy_statements__without(self, s3_bucket, key, value):
        statements = self.policy_statements(s3_bucket)
        new_list = []
        for statement in statements:
            if statement.get(key) != value:
                new_list.append(statement)
        return new_list

    def policy_create(self, s3_bucket, statements):
        #if (type(statements) is str) is False:
        #    statements = json.dumps(statements)

        policy = json.dumps({ 'Statement': statements   ,
                            'Version'  : '2012-10-17' })
        return self.s3().put_bucket_policy(Bucket= s3_bucket, Policy=policy)
        #return policy

    def transfer_config(self):
        return TransferConfig(use_threads=self.use_threads)


from unittest import TestCase

from osbot_aws.aws.ec2.EC2 import EC2
from osbot_aws.aws.ec2.utils.EC2__with_temp_role import EC2__with_temp_role


class TestCase__EC2(TestCase):
    ec2        : EC2

    @classmethod
    def setUpClass(cls):
        cls.ec2      = EC2__with_temp_role()

    @classmethod
    def tearDownClass(cls):
        pass


import json

from osbot_utils.utils import Misc

from osbot_aws.apis.Session import Session


class Parameter:                                # todo refactor main methods to SSM class
    def __init__(self, name=None):
        self._ssm = None
        self.name = name

    # helpers
    def ssm(self):
        if self._ssm is None:
            self._ssm = Session().client('ssm')
        return self._ssm

    # methods

    def create(self,value):
        return self.put(value)

    def delete(self):
        if self.exists() is False: return False
        self.ssm().delete_parameter(Name=self.name)
        return self.exists() is False

    def exists(self):

        return self.get() is not None

    def get(self):
        try:
            return self.ssm().get_parameter(Name=self.name).get('Parameter')
        except:
            return None

    def get_secret(self):
        try:
            return self.ssm().get_parameter(Name=self.name,WithDecryption=True).get('Parameter')
        except:
            return None

    def list(self):
        params = {}
        for param in self.ssm().describe_parameters().get('Parameters'):
            params[param.get('Name')] = param
        return params

    def push(self,data):
        return self.put(json.dumps(data))

    def pull(self):
        try:
            return json.loads(self.value())
        except:
            return None

    def push_secret(self,data):
        return self.put_secret(json.dumps(data))

    def pull_secret(self):
        try:
            return json.loads(self.value_secret())
        except:
            return None

    def put(self, value, description='', type='String',overwrite=True):
        return self.ssm().put_parameter(Name=self.name, Value=value, Description=description,Type=type, Overwrite=overwrite)

    def put_secret(self,value,description=''):
        return self.put(value,description=description,type='SecureString')

    def value(self):
        return Misc.get_value(self.get(),'Value')

    def value_secret(self):
        return Misc.get_value(self.get_secret(),'Value')

    def set_name(self,value): self.name = value ; return self


from osbot_aws.apis.test_helpers.Temp_Event_Rule import Temp_Event_Rule
from osbot_aws.apis.test_helpers.Temp_SQS_Queue import Temp_SQS_Queue


class Temp_Event_Rule_To_SQS_Queue:

    def __init__(self, event_source=None):
        self.sqs_queue    = None
        self.event_rule   = None
        self.target_id    = None
        self.event_source = event_source

    def __enter__(self):
        self.temp_sqs_queue  = Temp_SQS_Queue()
        self.temp_event_rule = Temp_Event_Rule(event_source=self.event_source)
        self.sqs_queue       = self.temp_sqs_queue.__enter__()
        self.event_rule      = self.temp_event_rule.__enter__()
        self.target_id       = self.event_rule.add_target_sqs_queue(self.sqs_queue)
        return self

    def __exit__(self, type, value, traceback):
        self.event_rule.delete_target(target_id=self.target_id)
        self.temp_sqs_queue .__exit__(type, value, traceback)
        self.temp_event_rule.__exit__(type, value, traceback)


from osbot_utils.decorators.lists.index_by import index_by

from osbot_utils.decorators.methods.cache import cache
from osbot_utils.decorators.methods.cache_on_self import cache_on_self

from osbot_aws.apis.Session import Session
from osbot_utils.utils.Misc import base64_to_str


class ECR:
    def __init__(self):
        pass

    @cache_on_self
    def client(self):
        return Session().client('ecr')

    def authorization_token(self):
        auth_data          = self.client().get_authorization_token().get('authorizationData')[0]
        auth_token         = base64_to_str(auth_data.get('authorizationToken'))
        proxy_endpoint     = auth_data.get('proxyEndpoint')
        username, password = auth_token.split(':')
        return {    "registry": proxy_endpoint  ,
                    "username": username        ,
                    "password": password        }

    def images(self, repository_name):
        result = self.client().describe_images(repositoryName=repository_name)
        return result.get('imageDetails')

    @index_by
    def images_ids(self, repository_name):
        result = self.client().list_images(repositoryName=repository_name)
        return result.get('imageIds')

    def registry(self):
        response = self.client().describe_registry()
        return { 'registry_id'               : response.get('registryId'              ),
                 'replication_configuration' : response.get('replicationConfiguration')}

    def repository_create(self, name, tags=None):
        if self.repository_exists(name):
            return { 'status': 'warning', 'message': f'repository {name} already existed'}
        kwargs = { 'repositoryName' : name,
                   'tags'           :  [] }
        if tags:
            for key,value in tags.items():
                kwargs.get('tags').append({'Key': key , 'Value':value})

        result = self.client().create_repository(**kwargs)
        return result.get('repository')

    def repository_delete(self, name):
        if self.repository_exists(name):
            self.client().delete_repository(repositoryName=name)
            return True
        return False

    def repository_exists(self, name):
        return self.repository_info(name) is not None

    def repository_info(self, name):
        try:
            result = self.client().describe_repositories(repositoryNames=[name])
            return result.get('repositories')[0]                                    # there should only be one repository here
        except:                                                                     # todo catch the actual Exception raised
            return None                                                             # when repository doesn't exist

    @index_by
    def repositories(self):
        response  = self.client().describe_repositories()
        return response.get('repositories')



from osbot_aws.aws.s3.S3                            import S3
from osbot_aws.aws.s3.S3__with_temp_role            import S3__with_temp_role
from osbot_aws.testing.Pytest                       import skip_pytest___aws_pytest_user_name__is_not_set
from osbot_aws.testing.TestCase__Boto3_Cache        import TestCase__Boto3_Cache
from osbot_utils.base_classes.Type_Safe             import Type_Safe


class TestCase__S3(TestCase__Boto3_Cache, Type_Safe):
    type_S3          : type
    client__s3       : object
    s3               : S3
    reset_iam_creds : bool = False

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        skip_pytest___aws_pytest_user_name__is_not_set()
        cls.type_S3           = S3__with_temp_role
        cls.s3                = cls.type_S3()
        cls.client__s3        = cls.s3.client

        if cls.reset_iam_creds:
            cls.s3       .temp_role__iam_reset_credentials()


from osbot_aws.aws.iam.IAM import IAM
from osbot_utils.utils.Http import GET
from osbot_aws.apis.API_Gateway import API_Gateway


class Rest_API:
    def __init__(self, api_name=None, api_id=None):
        self.api_name    = api_name
        self.api_id      = api_id
        self.api_gateway = API_Gateway()

    def create(self):
        result      = self.api_gateway.rest_api_create(self.api_name)
        self.api_id = result['id']
        return self

    def delete(self):
        self.api_gateway.rest_api_delete(self.id())
        self.api_id = None
        return self

    def deploy(self, stage='prod'):
        return self.api_gateway.deployment_create(self.id(),stage)

    def id(self):
        if self.api_id is None:
            self.api_id = self.api_gateway.rest_api_id(self.api_name)
        return self.api_id

    def exists(self):
        return self.id() is not None

    def add_method_http(self, from_path, from_method, to_method, to_uri):
        resource_id     = self.resource_id(from_path)
        status_code     = '200'
        response_models = {'application/json': 'Empty'}
        response_templates = {'application/json': ''}
        method_create               = self.api_gateway.method_create              (self.id() , resource_id , from_method                                   )
        integration_create__http    = self.api_gateway.integration_create__http   (self.id() , resource_id , to_uri      , from_method , to_method         )
        method_response_create      = self.api_gateway.method_response_create     (self.id() , resource_id , from_method , status_code ,response_models    )
        integration_response_create = self.api_gateway.integration_response_create(self.id() , resource_id , from_method , status_code , response_templates)
        return { 'method_create'              : method_create               ,
                 'integration_create__http'   : integration_create__http    ,
                 'method_response_create'     : method_response_create      ,
                 'integration_response_create': integration_response_create }

    def add_method_lambda(self, from_path, from_method, lambda_name):
        resource_id        = self.resource_id(from_path)
        status_code        = '200'
        response_models    = {'application/json': 'Empty'}
        response_templates = {'application/json': ''}
        method_create               = self.api_gateway.method_create(self.api_id, resource_id,from_method)
        integration_create__lambda  = self.api_gateway.integration_create__lambda(self.id(), resource_id, lambda_name,from_method)
        integration_add_permission  = self.api_gateway.integration_add_permission_to_lambda(self.id(), lambda_name)
        method_response_create      = self.api_gateway.method_response_create(self.id(),resource_id,from_method, status_code,response_models)
        integration_response_create = self.api_gateway.integration_response_create(self.id(),resource_id, from_method,status_code, response_templates)
        return { 'method_create'              : method_create               ,
                 'integration_create__lambda' : integration_create__lambda    ,
                 'integration_add_permission' : integration_add_permission,
                 'method_response_create'     : method_response_create      ,
                 'integration_response_create': integration_response_create }


    def method(self, path, method):
        resource_id = self.resource_id(path)
        return self.api_gateway.method(self.id(), resource_id, method)

    def not_exists(self):
        return self.exists() is False

    def resource_id(self, path):
        return self.api_gateway.resource(self.id(), path).get('id')

    def test_method(self, path='/',method='GET'):
        return self.api_gateway.method_invoke_test(self.id(),self.resource_id(path), method)

    def url(self,path='', stage='prod'):
        region = IAM().region()
        return self.api_gateway.stage_url(self.id(), region, stage, resource=path)

    def invoke_GET(self, path='', stage='prod'):
        return GET(self.url(path,stage))


from functools import cache

from dotenv import load_dotenv

from osbot_aws.aws.dynamo_db.Dynamo_DB import Dynamo_DB
from osbot_aws.aws.iam.IAM_Assume_Role import IAM_Assume_Role


class Dynamo_DB__with_temp_role(Dynamo_DB):
    temp_role__service  : str = "dynamodb"
    temp_role__action   : str = "*"
    temp_role__resource : str = "*"
    temp_role__name     : str = 'osbot__temp_role_for__test_Dynamo_DB'
    temp_role__recreate : bool = False

    
    @cache
    def client(self):
        load_dotenv()
        service       = self.temp_role__service
        action        = self.temp_role__action
        resource      = self.temp_role__resource
        role_name     = self.temp_role__name
        role_recreate = self.temp_role__recreate
        policies_to_add = [dict(service=service, action=action, resource=resource)]
        iam_assume_role = IAM_Assume_Role(role_name=role_name, policies_to_add=policies_to_add)
        iam_assume_role.create_role(recreate=role_recreate)
        #iam_assume_role.credentials_reset()
        return iam_assume_role.boto3_client(service_name=service)

    def temp_role__iam_reset_credentials(self):
        iam_assume_role = IAM_Assume_Role(role_name=self.temp_role__name)
        return iam_assume_role.credentials_reset()


from enum import Enum, auto

from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self

# class Sqlite__Bedrock__Row__Type:
#     UNIT_TEST   = 'unit_test'
#     DEV_REQUEST = 'dev_request'
#     CI_REQUEST  = 'ci_request'
#     UNDEFINED   = 'undefined'


class Sqlite__Bedrock__Row(Kwargs_To_Self):
    request_hash  : str
    request_data  : str
    response_hash : str
    response_data : str
    cache_hits    : int
    timestamp     : int
    latest        : bool
    comments      : str
    #record_type   : str  = Sqlite__Bedrock__Row__Type.UNDEFINED     # todo: see if this is really needed


import  logging
from    unittest                import TestCase
from    osbot_utils.utils       import Png
from    osbot_utils.utils.Dev   import Dev, pprint



class Test_Helper(TestCase):

    def setUp(self): #-> OSBot_Setup:
        logging.getLogger().addHandler(logging.StreamHandler())
        self.result     = None
        self.png_data   = None
        self.png_file   = '/tmp/unit-test.png'
        print()
        #return self.osbot_setup()

    #def osbot_setup(self,profile_name = None, account_id=None, region=None) -> OSBot_Setup:
        #return OSBot_Setup(profile_name=profile_name, account_id=account_id, region_name=region)

    def tearDown(self):
        if hasattr(self, 'result'):
            if self.result is not None:
                pprint(self.result)
        if hasattr(self, 'png_data'):
            if hasattr(self, 'png_data') is False:
                self.png_file = '/tmp/unit-test.png'
            self.save_png(self.png_data, self.png_file)
        #else:
        #    pprint("**** TEST HELPER WARNING - super().setUp() was not called from the current setUp override method **** ")

    # def lambda_package(self, lambda_name, profile_name = None, account_id=None, region=None):
    #     return self.osbot_setup(profile_name=profile_name,account_id=account_id,region=region).lambda_package(lambda_name)

    @staticmethod
    def print(result):
        if result is not None:
            pprint(result)

    @staticmethod
    def save_png(png_data, target_file):
        Png.save_png_base64_to_file(png_data, target_file)



from osbot_aws.apis.Lambda import Lambda
from osbot_aws.aws.iam.IAM import IAM


class IAM_Utils:            # todo see if there is a better home for the methods below

    def __init__(self):
        self.arn_aws_policy_service_sqs_lambda = 'arn:aws:iam::aws:policy/service-role/AWSLambdaSQSQueueExecutionRole'

    def policy_add_sqs_permissions_to_lambda_role(self, lambda_name):
        aws_lambda    = Lambda(lambda_name)
        iam_role_name = aws_lambda.configuration().get('Role').split(':role/').pop()
        IAM(role_name=iam_role_name).role_policy_attach(self.arn_aws_policy_service_sqs_lambda)
        return iam_role_name

    def policy_remove_sqs_permissions_to_lambda_role(self, lambda_name):
        aws_lambda    = Lambda(lambda_name)
        iam_role_name = aws_lambda.configuration().get('Role').split(':role/').pop()
        IAM(role_name=iam_role_name).role_policy_detach(self.arn_aws_policy_service_sqs_lambda)
        return iam_role_name


# from functools import wraps
#
#
# class aws_retry:
#
#     """retries the call (required for some cases where data is not consistent yet in AWS"""
#     def __init__(self, fields):
#         self.fields = fields                                        # field to inject
#
#     def __call__(self, function):
#         pass
#         #code from aws_inject
#         # from osbot_aws.AWS_Config import AWS_Config
#         # @wraps(function)                                            # makes __name__ work ok
#         # def wrapper(*args,**kwargs):                                # wrapper function
#         #     for field in self.fields.split(','):                    # split value provided by comma
#         #         if field == 'region'    : kwargs[field] = AWS_Config().aws_session_region_name()
#         #         if field == 'account_id': kwargs[field] = AWS_Config().aws_session_account_id()
#         #     return function(*args,**kwargs)
#         #return wrapper


from osbot_aws.aws.bedrock.models.claude.Anthropic__Claude_Instant_V1 import Anthropic__Claude_Instant_V1


class Anthropic__Claude_V2_0(Anthropic__Claude_Instant_V1):

    model_id = 'anthropic.claude-v2'


from osbot_aws.AWS_Config               import AWS_Config
from osbot_aws.aws.s3.S3                import S3
from osbot_aws.helpers.Lambda_Package   import Lambda_Package

# todo: add support for running these tests in different accounts, but for now use this to make sure we are always in the correct AWS account
CURRENT__OSBOT_AWS__TESTS__ACCOUNT_ID     = '381492182978'
CURRENT__OSBOT_AWS__TESTS__IAM_USER       = 'OSBot-AWS-Dev__Only-IAM'
CURRENT__OSBOT_AWS__TESTS__DEFAULT_REGION = 'eu-west-1'

class OSBot_Setup:      # todo refactor out this code, most of this has been moved into aws_config and Deploy_Lambda

    def __init__(self, bot_name= None, profile_name = None, account_id=None, region_name=None, lambda_s3_bucket=None, lambda_role_name=None):
        self.aws_config        = AWS_Config()
        if bot_name            : self.aws_config.set_bot_name                 (bot_name)
        if profile_name        : self.aws_config.set_aws_session_profile_name(profile_name     )
        if account_id          : self.aws_config.set_aws_session_account_id   (account_id      )
        if region_name         : self.aws_config.set_aws_session_region_name  (region_name     )
        if lambda_s3_bucket    : self.aws_config.set_lambda_s3_bucket         (lambda_s3_bucket)
        if lambda_role_name    : self.aws_config.set_lambda_role_name         (lambda_role_name)

        self.bot_name          = self.aws_config.bot_name()
        self.profile_name      = self.aws_config.aws_session_profile_name()
        self.region_name       = self.aws_config.aws_session_region_name()
        self.account_id        = self.aws_config.aws_session_account_id()
        self.s3_bucket_lambdas = self.aws_config.lambda_s3_bucket()
        self.lambda_role_name  = self.aws_config.lambda_role_name()
        self.lambda_role_arn   = f"arn:aws:iam::{self.account_id}:role/{self.lambda_role_name}"

        self.s3                = S3()

    def lambda_package(self, lambda_name) -> Lambda_Package:
        lambda_package               = Lambda_Package(lambda_name)
        lambda_package.tmp_s3_bucket = self.s3_bucket_lambdas                       # these four method calls need to be refactored
        lambda_package.tmp_s3_key    = 'lambdas/{0}.zip'.format(lambda_name)
        lambda_package.aws_lambda.set_s3_bucket(lambda_package.tmp_s3_bucket)
        lambda_package.aws_lambda.set_s3_key(lambda_package.tmp_s3_key)
        return lambda_package

    # def setup_test_environment(self):
    #     Globals.aws_session_profile_name = self.profile_name
    #     Globals.aws_session_region_name  = self.region_name
    #     return self

    def set_up_buckets(self):
        if self.s3_bucket_lambdas not in self.s3.buckets():                         # todo: refactor to use #s3.bucket_exists()
            result = self.s3.bucket_create(self.s3_bucket_lambdas,self.region_name)
            assert result.get('status') == 'ok'
        return self






import jwt
import requests

from osbot_utils.decorators.methods.remove_return_value import remove_return_value

from osbot_utils.utils.Misc import datetime_to_str

from osbot_utils.decorators.lists.group_by import group_by

from osbot_utils.decorators.lists.index_by import index_by

from osbot_utils.decorators.methods.cache_on_self import cache_on_self

from osbot_aws.apis.Session import Session


class Cognito_IDP:

    def __init__(self):
        pass

    @cache_on_self
    def cognito(self):
        return Session().client('cognito-idp')

    @remove_return_value(field_name='ResponseMetadata')
    def auth_initiate(self, client_id, username, password):

        auth_parameters = { 'USERNAME': username                ,
                            'PASSWORD': password                ,
                            'SCOPE'   : 'openid profile email'  }
        client = self.cognito()
        try:
            response              = client.initiate_auth( ClientId=client_id, AuthFlow='USER_PASSWORD_AUTH',AuthParameters=auth_parameters)
            access_token          = response.get('AuthenticationResult').get('AccessToken')
            return {'auth_result'     : response.get('AuthenticationResult'),
                    'challenge_params': response.get('ChallengeParameters' ),
                    'jwt_token'       : self.decode_jwt_token(access_token )}

        except Exception as error:
            return {'error':  str(error) }

    #todo: find a workaround since this is broken in AWS (see https://stackoverflow.com/questions/52425678/access-token-does-not-contain-openid-scope-in-aws-cognito)
    def auth_user_info(self, project, region, access_token):
        user_info_url = f"https://{project}.auth.{region}.amazoncognito.com/oauth2/userInfo"

        headers = { 'Authorization': f"Bearer {access_token}" }
        response = requests.get(user_info_url, headers=headers)
        return response.json()

    def decode_jwt_token(self, access_token):
        if access_token:
            return jwt.decode(access_token, algorithms=["RS256"], options={"verify_signature": False})

    def user_info(self,  user_pool_id, user_name):
        return self.user(user_pool_id, user_name)

    def user(self, user_pool_id, user_name):
        client = self.cognito()
        try:
            response = client.admin_get_user(UserPoolId=user_pool_id, Username=user_name)
            return self.format_user_data(response)
        except client.exceptions.ResourceNotFoundException:
            return {}
        except client.exceptions.UserNotFoundException:
            return {}


    def user_create(self, user_pool_id, user_name, user_attributes, temporary_password):
        if self.user_exists(user_pool_id, user_name):
            return {}
        kwargs = dict(UserPoolId        = user_pool_id      ,
                      Username          = user_name         ,
                      UserAttributes    = user_attributes   ,
                      TemporaryPassword = temporary_password,
                      #DesiredDeliveryMediums = []          # not working
                      )
        response = self.cognito().admin_create_user(**kwargs)
        return self.format_user_data(response.get('User'))

    def user_delete(self, user_pool_id, user_name):
        if self.user_exists(user_pool_id, user_name):
            client   = self.cognito()
            client.admin_delete_user(UserPoolId=user_pool_id, Username=user_name)
            return True
        return False

    def user_exists(self, user_pool_id, user_name):
        return self.user(user_pool_id, user_name) != {}

    def user_pool(self, user_pool_id):
        response  = self.cognito().describe_user_pool(UserPoolId=user_pool_id)
        user_pool = response.get('UserPool')
        schema    = {}
        for attribute in user_pool.get('SchemaAttributes'):
            schema[attribute.get('Name')] = attribute.get('AttributeDataType')

        data = { 'arn'             : user_pool.get('Arn'                   ),
                 'domain'          : user_pool.get('Domain'                ),
                 'id'              : user_pool.get('Id'                    ),
                 'user_count'      : user_pool.get('EstimatedNumberOfUsers'),
                 'name'            : user_pool.get('Name'                  ),
                 'password_policy' : user_pool.get('Policies',{}           ).get('PasswordPolicy',{}),
                 'schema'          : schema                                 ,
                 'status'          : user_pool.get('Status'                ),
                 'tags'            : user_pool.get('UserPoolTags'          )}
        return data


    def user_set_password(self, user_pool_id, user_name, password, permanent=True):
        if self.user_exists(user_pool_id, user_name):
            client   = self.cognito()
            try:
                kwargs   = dict(UserPoolId=user_pool_id, Username=user_name, Password=password, Permanent=True)
                client.admin_set_user_password(**kwargs)
                return True
            except client.exceptions.InvalidParameterException:
                return False
        return False

    @index_by
    @group_by
    def users(self, user_pool_id, limit=10):
        response = self.cognito().list_users( UserPoolId=user_pool_id, Limit=limit )
        users = []
        for user_data in response.get('Users'):
            user = self.format_user_data(user_data)
            users.append(user)
        return users

    def format_user_data(self, user_data):
        user = {'user_status'            : user_data.get('UserStatus'                          , None),
                'enabled'                : user_data.get('Enabled'                             , None),
                'user_create_date'       : datetime_to_str(user_data.get('UserCreateDate'      , None)),
                'user_last_modified_date': datetime_to_str(user_data.get('UserLastModifiedDate', None)),
                'username'               : user_data.get('Username'                            , None)}

        for attribute in user_data.get('UserAttributes', []):
            name       = attribute['Name']
            value      = attribute['Value']
            user[name] = value
        return user




from osbot_utils.utils.Files import path_combine, current_temp_folder, folder_sub_folders, folders_names


class Lambda_Layers_Local:
    DEFAULT_TARGET_FOLDER_NAME = '_lambda_dependencies'

    def __init__(self):
        pass

    def path_lambda_dependencies(self):
        return path_combine(current_temp_folder(), self.DEFAULT_TARGET_FOLDER_NAME)

    def installed_layers(self):
        folders = folder_sub_folders(self.path_lambda_dependencies())
        return folders_names(folders)



from osbot_utils.decorators.lists.index_by import index_by
from osbot_utils.decorators.methods.catch import catch

from osbot_aws.apis.Session import Session


class ACM:

    def __init__(self):
        self._acm       = None

    def acm(self):
        if self._acm is None:
            self._acm = Session().client('acm')
        return self._acm

    @catch
    def certificate(self, certificate_arn, include_certs=False):
        cert_info = self.acm().describe_certificate(CertificateArn=certificate_arn).get('Certificate')
        if include_certs:
            cert_data = self.acm().get_certificate(CertificateArn=certificate_arn)
            cert_info['Certificate'     ] = cert_data.get('Certificate')
            cert_info['CertificateChain'] = cert_data.get('CertificateChain')

        return cert_info

    @catch
    def certificate_request(self, domain_name, validation_method='DNS'):
        return self.acm().request_certificate(DomainName=domain_name, ValidationMethod=validation_method)


    @catch
    @index_by
    def certificates(self):
        return self.acm().list_certificates().get('CertificateSummaryList')


import json
from functools import cache

import boto3

from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Objects import get_value

from osbot_utils.decorators.methods.cache_on_self   import cache_on_self
from osbot_utils.decorators.lists.group_by          import group_by
from osbot_utils.decorators.lists.index_by          import index_by
from osbot_utils.decorators.methods.catch           import catch
from osbot_utils.decorators.methods.remove_return_value import remove_return_value
from osbot_utils.testing.Duration import Duration
from osbot_utils.utils.Json import json_to_str, from_json_str
from osbot_utils.utils.Misc import random_string, wait
from osbot_utils.utils.Status import status_ok, status_error

from osbot_aws.apis.Session                         import Session
from osbot_aws.aws.sts.STS import STS


class IAM:

    # todo: remove user_name and role_name from here
    def __init__(self, user_name=None, role_name=None):
        self.user_name   = user_name                        # todo: refactor this variable into the method's definition (to make it consistent with the other APIs)
        self.role_name   = role_name                        #       this will be a bit of a breaking change to code that uses this class for role management

    # helpers
    @cache
    def client(self):                                          # rename to client
        return Session().client('iam')

    @cache
    def resource(self):
        return Session().resource('iam')

    @cache
    def session(self):
        return Session().session()

    @cache
    def sts(self):
        return STS()


    # main method
    #todo: refactor with better wait solution
    def access_key__wait_until_key_is_working(self, access_key, wait_for=0.5 , wait_count=20, success_count=5):
        """
        Wait until the access key is working. When testing this there a number of cases where
        access_key__is_key_working would return False after a True one (which means that AWS was still not fully syncronised (see https://aws.amazon.com/iam/faqs/ and https://stackoverflow.com/questions/20156043/how-long-should-i-wait-after-applying-an-aws-iam-policy-before-it-is-valid)
        :param access_key:
        :param wait_count:
        :param success_count:  if still seeing some false positives include the default value (set to 5)
        :return:
        """
        if success_count < 1:                                                   # make sure this value is not below 1
            success_count = 1
        status    = False
        key_is_ok = 0
        with Duration() as duration:
            for i in range(1,wait_count):
                if self.access_key__is_key_working(access_key) is True:         # the get_caller_identity takes about 500ms to execute
                    key_is_ok += 1                                              # increment key_is_ok
                    if key_is_ok == success_count:                              # only after success_count matches, we can reliable say that key is good
                        status = True
                        break
                    continue
                wait(wait_for)
        if status:
            message = f'access key ok: during {duration.seconds()} seconds, confirmed {key_is_ok} times that the key is working'
            return status_ok(message=message)
        else:
            message = f'access not working ok: after {duration.seconds()} seconds, was only able to confirm {key_is_ok} times that the key was working'
            return status_error(message=message)

    def access_key__wait_until_key_is_not_working(self, access_key, wait_count=100, success_count=30):
        """
        waits until the access key is not working anymore
        note: I couldn't get this to provide a consistent result with values smaller than 30
        :param access_key:
        :param wait_count:
        :param success_count:
        :return:
        """

        for i in range(1,wait_count):
            if self.access_key__is_key_working(access_key) is False:        # the get_caller_identity takes about 500ms to execute
                success_count -= 1
                if success_count == 0:
                    print(f'waited {i} times - before confirming key is not working anymore')
                    return True
        return False

    def access_key__is_key_working(self, access_key):       # todo refactor to Sessions class
        aws_access_key_id     = access_key.get('AccessKeyId')
        aws_secret_access_key = access_key.get('SecretAccessKey')
        try:
            session = boto3.Session(aws_access_key_id=aws_access_key_id, aws_secret_access_key=aws_secret_access_key)
            session.client('sts').get_caller_identity()
            return True
        except Exception as error:
            assert error.__str__() == 'An error occurred (InvalidClientTokenId) when calling the GetCallerIdentity operation: The security token included in the request is invalid.'
            return False

    @group_by
    @index_by
    def access_keys(self):
        return self.client().list_access_keys().get('AccessKeyMetadata')

    @cache
    def account_id(self):
        return self.sts().current_account_id()

    def check_aws_security_tokens(self):
        try:
            identity = self.caller_identity()
            return {'status': "Ok", 'error': None, "data": identity}
        except Exception as error:
            return { 'status': "Error", 'error':error, "data":None}


    @cache_on_self
    def region(self):
        return self.session().region_name

    @remove_return_value('ResponseMetadata')
    def caller_identity(self):
        return self.sts().caller_identity()

    def get_data(self, method, field_id, use_paginator, **kwargs):
        paginator = self.client().get_paginator(method)
        for page in paginator.paginate(**kwargs):
            for id in page.get(field_id):
                yield id
            if use_paginator is False:
                return

    def groups(self):
        return self.get_data('list_groups', 'Groups', True)

    @catch
    def login_profile(self):
        return self.client().get_login_profile(UserName=self.user_name)

    def login_profile_create(self, password, reset_required=True):
        return self.client().create_login_profile(UserName=self.user_name, Password=password, PasswordResetRequired=reset_required)

    def login_profile_delete(self):
        if self.login_profile_exists():
            self.client().delete_login_profile(UserName=self.user_name)

    def login_profile_exists(self):
        return self.login_profile().get('error') is None

    def policy_arn(self, policy_name, policy_path='/',account_id=None):
        if policy_name is None: return policy_name
        if account_id  is None: account_id = self.account_id()
        if policy_path is None or policy_path == '/':
            return f'arn:aws:iam::{account_id}:policy/{policy_name}'
        return f'arn:aws:iam::{account_id}:policy{policy_path}/{policy_name}'

    def policy_create_version(self, policy_name, policy_document, policy_path='/', account_id=None): # todo: add support for deleting previous versions
        policy_arn = self.policy_arn(policy_name, policy_path, account_id)
        if type(policy_document) is not str:
            policy_document = json.dumps(policy_document)
        data = self.client().create_policy_version(PolicyArn=policy_arn, PolicyDocument=policy_document, SetAsDefault=True).get('Policy')
        return {'status': 'ok', 'policy_name': policy_name, 'policy_arn': policy_arn, 'data': data}

    def policies_create(self, policies, project_name=None, recreate_policy=False):
        policies_arns           = []
        existing_policies       = self.role_policies()
        existing_policies_names = list(existing_policies.keys())
        project_name            = project_name or random_string()
        for base_name, policy in policies.items():
            policy_name = "{0}_{1}".format(base_name, project_name)
            if policy_name in existing_policies_names:
                existing_policy_arn = existing_policies[policy_name]
                if recreate_policy:
                    self.policy_delete(existing_policy_arn)
                else:
                    policies_arns.append(existing_policy_arn)
                    continue
            result = self.policy_create(policy_name, policy)
            policies_arns.append(result.get('policy_arn'))
        return policies_arns

    def policy_create(self, policy_name, policy_document, policy_path='/', account_id=None, delete_before_create=False):
        policy_arn = self.policy_arn(policy_name, policy_path,account_id)
        if delete_before_create:
            self.policy_delete(policy_arn)
        if self.policy_exists(policy_arn) is False:
            if type(policy_document) is not str:
                policy_document = json.dumps(policy_document)
            try:
                data       = self.client().create_policy(PolicyName=policy_name, PolicyDocument=policy_document).get('Policy')
                policy_arn = data.get('Arn')
                return {'status': 'ok'     , 'policy_name': policy_name, 'policy_arn': policy_arn                  , 'data': data                     }
            except Exception as error:
                return {'status': 'error'  , 'policy_name': policy_name, 'policy_arn': None                        , 'data': '{0}'.format(error)      }
        return {        'status': 'warning', 'policy_name': policy_name, 'policy_arn': policy_arn                  , 'data': 'policy already existed' }

    def policy_delete(self, policy_arn=None, policy_name=None):
        if policy_arn:
            if self.policy_exists(policy_arn) is False: return False
        elif policy_name:
            role_policies = self.role_policies()
            if policy_name in role_policies:
                policy_arn=role_policies.get(policy_name)
            else:
                return False
        else:
            return False
        self.policy_detach_roles(policy_arn)
        self.client().delete_policy(PolicyArn= policy_arn)
        return self.policy_exists(policy_arn) is False

    def policy_delete_by_name(self, policy_name, policy_path='/',account_id=None):
        policy_arn = self.policy_arn(policy_name,policy_path,account_id)
        if self.policy_exists(policy_arn) is False: return False
        self.client().delete_policy(PolicyArn= policy_arn)
        return self.policy_exists(policy_arn) is False

    def policy_info(self, policy_arn):
        try:
            return self.client().get_policy(PolicyArn=policy_arn).get('Policy')
        except:
            return None

    def policy_inline_delete(self, policy_name):
        return self.client().delete_role_policy(RoleName=self.role_name, PolicyName=policy_name)

    def policy_details(self, policy_arn):
        try:
            policy_info    = self.policy_info(policy_arn)
            if policy_info:
                policy_arn     = policy_info.get('Arn')
                version_id     = policy_info.get('DefaultVersionId')
                policy_version = self.client().get_policy_version(PolicyArn=policy_arn, VersionId=version_id).get('PolicyVersion')
                return {'policy_arn': policy_arn, 'policy_name': policy_info.get('PolicyName'), 'policy_info' : policy_info, 'policy_version': policy_version}
        except:
            return None

    def policy_detach_roles(self,policy_arn):
        policy_roles = self.client().list_entities_for_policy(PolicyArn=policy_arn).get('PolicyRoles')
        for role in policy_roles:
            IAM(role_name=role.get('RoleName')).role_policy_detach(policy_arn)
        return self

    def policy_statement(self, policy_arn):
        policy_details = self.policy_details(policy_arn)
        if policy_details:
            return policy_details.get('policy_version').get('Document').get('Statement')

    def policy_exists(self, policy_arn):
        return self.policy_info(policy_arn) is not None

    def policy_exists_by_name(self, policy_name, policy_path='/', account_id=None):
        calculated_policy_arn = self.policy_arn(policy_name, policy_path, account_id)   # since there doesn't seem to be a way to search for a policy by name
        return self.policy_info(calculated_policy_arn) is not None

    def policy_not_exists(self, policy_arn):
        return self.policy_exists(policy_arn) is False

    @index_by
    def policies(self):
        return self.get_data('list_policies', 'Policies', True)

    def policies_delete(self, policies_names):
        for policy_name in policies_names:
            self.policy_delete(policy_name)

    def role_arn(self):
        return get_value(self.role_info(), 'Arn')

    def role_assume_policy(self):
        return self.role_info().get('AssumeRolePolicyDocument')

    def role_assume_policy_update(self, policy_document):
        if type(policy_document) is not str:
            policy_document = json_to_str(policy_document)
        return self.client().update_assume_role_policy(RoleName=self.role_name, PolicyDocument=policy_document)

    def role_exists(self):
        return self.role_info() is not None

    def role_info(self):
        try:
            return self.client().get_role(RoleName=self.role_name).get('Role')
        except:
            return None

    def role_create(self, assume_policy_document, skip_if_exists=True):
        role_info = self.role_info()
        if role_info:
            if skip_if_exists:
                return role_info                     # todo: confirm that the values we get from self.role_info() are compatible with the values in .get('Role') (below)
            else:
                self.role_delete(check_exists=False)
        if type(assume_policy_document) is not str:
            assume_policy_document = json.dumps(assume_policy_document)
        result = self.client().create_role(RoleName=self.role_name, AssumeRolePolicyDocument=assume_policy_document)
        if result:
            return result.get('Role')

    def role_delete(self, check_exists=True):
        if check_exists and self.role_exists() is False:
            return False                                            # make sure it exists
        self.role_policies_detach_and_delete()                      # delete all associated policies
        result = self.client().delete_role(RoleName=self.role_name)
        if result:
            return result.get('ResponseMetadata',{}).get('HTTPStatusCode') == 200
        return False

    def role_create_instance_profile(self, role_name):
        return self.client().create_instance_profile(InstanceProfileName=role_name)

    def role_add_to_instance_profile(self, role_name, instance_profile_name):
        return self.client().add_role_to_instance_profile(InstanceProfileName = instance_profile_name,
                                                          RoleName            = role_name            )
    def role_not_exists(self):
        return self.role_exists() is False

    def role_policies_detach_and_delete(self):
        attached_policies_arns = self.role_policies_attached().values()
        self.role_policies_detach(attached_policies_arns)
        self.role_policies_delete()
        #self.policies_delete     (policies_arns)       # was a BUG, we shouldn't be deleting top level policies here

    def role_policy_add(self, policy_name, policy_document):
        if type(policy_document) is not str:
            policy_document = json_to_str(policy_document)
        return self.client().put_role_policy(RoleName=self.role_name,PolicyName=policy_name,PolicyDocument=policy_document)

    def role_policy_attach(self,policy_arn):
        return self.client().attach_role_policy(RoleName=self.role_name, PolicyArn=policy_arn)

    def role_policy_detach(self, policy_arn):
        self.client().detach_role_policy(RoleName=self.role_name, PolicyArn=policy_arn)

    def role_policies_detach(self, policies_arn):
        for policy_arn in policies_arn:
            self.role_policy_detach(policy_arn)
        return self

    def role_policies_delete(self):
        inline_policies = self.client().list_role_policies(RoleName=self.role_name)
        for policy_name in inline_policies['PolicyNames']:
            self.client().delete_role_policy(RoleName=self.role_name, PolicyName=policy_name)
        return self


    def role_policies_attach(self, policies_arn):
        for policy_arn in policies_arn:
            self.role_policy_attach(policy_arn)
        return self

    @index_by
    def role_policies(self):
        policies_attached = self.role_policies_attached()
        policies_inline   = self.role_policies_inline  ()
        return {**policies_attached, **policies_inline}

    @index_by
    def role_policies_attached(self):
        policies = {}
        if self.role_name:
            policies_attached = self.get_data('list_attached_role_policies', 'AttachedPolicies', True, RoleName=self.role_name)
            if policies_attached:
                for item in policies_attached:
                    policies[item.get('PolicyName')] = item.get('PolicyArn')
        return policies

    @index_by
    def role_policies_inline(self, fetch_inline_policies=False):
        policies = {}
        if self.role_name:
            policies_inline = self.get_data('list_role_policies', 'PolicyNames', True, RoleName=self.role_name)
            if policies_inline:
                for item in policies_inline:
                    #policies[item.get('PolicyName')] = item.get('PolicyArn')
                    policies[item] = {'PolicyName': item}                           # list_role_policies only returns an arroy of names, not the actual policy
        if fetch_inline_policies:
            for policy_name in policies.keys():

                policy_document = self.client().get_role_policy(RoleName=self.role_name, PolicyName=policy_name)
                policies[policy_name] = policy_document.get('PolicyDocument')
                #print(f"Policy Name: {policy_name}")
                #print("Policy Document:", policy_document['PolicyDocument'])

        return policies

    def role_policies_statements(self, just_statements = False):
        if just_statements:
            statements = []
            for items in self.role_policies_statements(just_statements=False).values():     # todo: find a better way to do this, this pattern looks and feels weird
                statements.extend(items)
            return statements

        policy_statements = {}
        for policy_name, policy_arn in self.role_policies().items():
            policy_statements[policy_name] = self.policy_statement(policy_arn)
        return policy_statements

    @index_by
    def roles(self):
        return list(self.get_data('list_roles', 'Roles', True))

    @index_by
    def users(self):
        return self.get_data('list_users', 'Users', True)

    def users_by_username(self):
        return self.users(index_by='UserName')

    def user_access_key_create(self, wait_for_key_working=False):
        access_key = self.client().create_access_key(UserName=self.user_name).get('AccessKey')
        if wait_for_key_working:
            result = self.access_key__wait_until_key_is_working(access_key)
            if result.get('status') == 'ok':
                return None
        return access_key

    def user_access_key_delete(self, access_key_id):
        return self.client().delete_access_key(UserName=self.user_name, AccessKeyId=access_key_id)

    @index_by
    def user_access_keys(self):
        return self.client().list_access_keys(UserName=self.user_name).get('AccessKeyMetadata')

    def user_access_keys_delete_all(self):
        for access_key in self.user_access_keys():
            self.user_access_key_delete(access_key.get('AccessKeyId'))
        return len(self.user_access_keys()) == 0

    def user_attach_policy(self, policy_arn):
        return self.client().attach_user_policy(UserName=self.user_name, PolicyArn=policy_arn)

    def user_exists(self):
        return self.user_info().get('error') is None

    @catch
    def user_info(self):
        return self.client().get_user(UserName=self.user_name).get('User')

    def user_create(self):
        if self.user_exists() is False:
            return self.client().create_user(UserName=self.user_name).get('User')
        return self.user_info()

    def user_delete(self):
        if self.user_exists() is False: return False            # return False if user doesn't exist
        self.login_profile_delete()                             # delete user's profile
        self.user_access_keys_delete_all()                      # delete user's access keys
        self.client().delete_user(UserName=self.user_name)         # delete user
        return self.user_exists() is False                      # return True if user doesn't exist any more (which it shouldn't)

    def user_polices(self):
        return self.client().list_user_policies(UserName=self.user_name)

    def set_user_name (self, value): self.user_name   = value ;return self
    def set_role_name (self, value): self.role_name   = value; return self

    def wait_for_policy(self, policy_name, delay=None, max_attempts=None):
        waiter_name = 'policy_exists'
        arg_name    = 'PolicyArn'
        return self.wait_for_waiter(waiter_name, arg_name, policy_name, delay, max_attempts)

    def wait_for_role(self, role_name, delay=None, max_attempts=None):
        waiter_name = 'role_exists'
        arg_name    = 'RoleName'
        return self.wait_for_waiter(waiter_name, arg_name, role_name, delay, max_attempts)

    def wait_for_waiter(self, waiter_name, arg_name, arg_value, delay=None, max_attempts=None):
        kwargs = {
            arg_name : arg_value
        }
        kwargs['WaiterConfig'] = {
            'Delay'      : delay        or 1,               # 1  is default boto3 value for Delay
            'MaxAttempts': max_attempts or 20               # 20 is default boto3 value for Delay
        }

        waiter = self.client().get_waiter(waiter_name)
        waiter.wait(**kwargs)
        return waiter




from botocore.exceptions import ClientError
from osbot_utils.decorators.methods.cache_on_self import cache_on_self

from osbot_utils.decorators.methods.remove_return_value import remove_return_value

from osbot_utils.decorators.lists.index_by import index_by

from osbot_aws.apis.Session               import Session

from osbot_utils.decorators.methods.cache import cache


class Events:

    @cache_on_self
    def client(self):
        return Session().client('events')

    @remove_return_value('ResponseMetadata')
    def archive(self, archive_name):
        return self.client().describe_archive(ArchiveName=archive_name)

    @index_by
    def archives(self, event_source_arn=None, name_prefix=None):
        kwargs = {}
        if event_source_arn: kwargs['EventSourceArn'] = event_source_arn
        if name_prefix: kwargs['NamePrefix'] = name_prefix
        return self.client().list_archives(**kwargs).get('Archives')

    @remove_return_value('ResponseMetadata')
    def events_put(self, events):
        return self.client().put_events(Entries=events)


    @remove_return_value('ResponseMetadata')
    def rule(self, rule_name, event_bus_name=None):
        try:
            kwargs = {'Name': rule_name}
            if event_bus_name: kwargs['EventBusName'] = event_bus_name
            return self.client().describe_rule(**kwargs)
        except ClientError as e:                                                    # todo: refactor this pattern into a helper with statement (which receives as param the Error.Code value
            if e.response.get('Error', {}).get('Code') == 'ResourceNotFoundException':
                return None
            raise

    def rule_arn(self, rule_name, event_bus_name=None):
        rule = self.rule(rule_name=rule_name, event_bus_name=event_bus_name)
        if rule:
            return rule.get('Arn')

    @remove_return_value('ResponseMetadata')
    def rule_create(self, rule_name, event_source=None, event_pattern=None, schedule_expression=None, description=None, tags=None, event_bus_name=None):
        kwargs = {  "Name"          : rule_name                            ,
                    "Description"   : description or ''                    ,
                    "State"         : 'ENABLED'                            ,
                    "Tags"          : []                                   }

        if event_bus_name:
            kwargs['EventBusName'] = event_bus_name
        if event_pattern:
            kwargs["EventPattern"] = event_pattern
        if event_source:
            kwargs["EventPattern"] = f'{{ "source": ["{event_source}"] }}'
        if schedule_expression:
            kwargs["ScheduleExpression"] = schedule_expression

        if tags:
            for key,value in tags.items():
                kwargs['Tags'].append({'Key':key, 'Value':value})
        return self.client().put_rule(**kwargs).get('RuleArn')

    def rule_delete(self, rule_name , event_bus_name=None):
        if self.rule_not_exists(rule_name): return False
        kwargs = { "Name" :rule_name,
                   "Force": True    }
        if event_bus_name: kwargs['EventBusName'] = event_bus_name
        self.client().delete_rule(**kwargs)
        return self.rule_not_exists(rule_name)

    def rule_exists(self, rule_name, event_bus_name=None):
        return self.rule(rule_name=rule_name, event_bus_name=event_bus_name) is not None

    def rule_not_exists(self, rule_name, event_bus_name=None):
        return self.rule_exists(rule_name, event_bus_name=event_bus_name) is False

    @index_by
    def rules(self, name_prefix=None, event_bus_name=None):
        kwargs = {}
        if name_prefix   : kwargs['NamePrefix'  ] = name_prefix
        if event_bus_name: kwargs['EventBusName'] = event_bus_name
        return self.client().list_rules(**kwargs).get('Rules')

    def target(self, rule_name, target_id):
        return self.targets(rule_name, index_by='Id').get(target_id)

    @remove_return_value('ResponseMetadata')
    def target_create(self, rule_name, target_id, target_arn, target_attributes=None):
        target = {"Arn": target_arn ,
                  "Id" : target_id  }

        if target_attributes:
            target.update(target_attributes)

        kwargs = { "Rule"   : rule_name    ,
                   "Targets": [target]     }

        return self.client().put_targets(**kwargs)

    @remove_return_value('ResponseMetadata')
    def target_delete(self, rule_name, target_id, event_bus_name=None):
        kwargs = {"Rule" : rule_name   ,
                  "Ids"  : [target_id] ,
                  "Force": True        }
        if event_bus_name: kwargs['EventBusName'] = event_bus_name
        return self.client().remove_targets(**kwargs)

    def target_exists(self, rule_name, target_id):
        return self.target(rule_name=rule_name, target_id=target_id) is not None

    @index_by
    @remove_return_value('ResponseMetadata')
    def targets(self, rule_name, event_bus_name=None):
        kwargs = {'Rule': rule_name}
        if event_bus_name: kwargs['EventBusName'] = event_bus_name
        return self.client().list_targets_by_rule(**kwargs).get('Targets')


import os

from osbot_utils.base_classes.Type_Safe import Type_Safe
from osbot_utils.utils.Env import load_dotenv

DEFAULT__AWS_DEFAULT_REGION = 'eu-west-1'

class AWS_Config(Type_Safe):

    def __init__(self):
        super().__init__()
        load_dotenv()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass

    def aws_access_key_id           (self): return os.getenv('AWS_ACCESS_KEY_ID'              )
    def aws_secret_access_key       (self): return os.getenv('AWS_SECRET_ACCESS_KEY'          )
    def aws_session_profile_name    (self): return os.getenv('AWS_PROFILE_NAME'               )
    def aws_session_region_name     (self): return os.getenv('AWS_DEFAULT_REGION'             ) or DEFAULT__AWS_DEFAULT_REGION
    def aws_session_account_id      (self): return os.getenv('AWS_ACCOUNT_ID'                 ) or  self.sts__session_account_id()

    def dev_skip_aws_key_check      (self): return os.getenv('DEV_SKIP_AWS_KEY_CHECK'        , False              )     # use to not have the 500ms check that happens during this check
    def bot_name                    (self): return os.getenv('OSBOT_NAME'                                         )     # todo: refactor variable to osbot_name (need to check for side effects)
    def lambda_s3_folder_layers     (self): return os.getenv('OSBOT_LAMBDA_S3_FOLDER_LAYERS' , 'layers'           )     # todo: change these static values to DEFAULT_.... ones
    def lambda_s3_folder_lambdas    (self): return os.getenv('OSBOT_LAMBDA_S3_FOLDER_LAMBDAS', 'lambdas'          )
    def lambda_role_name            (self): return os.getenv('OSBOT_LAMBDA_ROLE_NAME'          'role-osbot-lambda')

    def set_aws_access_key_id       (self, value): os.environ['AWS_ACCESS_KEY_ID'               ] = value ; return value
    def set_aws_secret_access_key   (self, value): os.environ['AWS_SECRET_ACCESS_KEY'           ] = value ; return value
    def set_aws_session_profile_name(self, value): os.environ['AWS_PROFILE_NAME'                ] = value ; return value
    def set_aws_session_region_name (self, value): os.environ['AWS_DEFAULT_REGION'              ] = value ; return value
    def set_aws_session_account_id  (self, value): os.environ['AWS_ACCOUNT_ID'                  ] = value ; return value
    def set_lambda_s3_bucket        (self, value): os.environ['OSBOT_LAMBDA_S3_BUCKET'          ] = value ; return value
    def set_lambda_s3_folder_layers (self, value): os.environ['OSBOT_LAMBDA_S3_FOLDER_LAYERS'   ] = value ; return value
    def set_lambda_s3_folder_lambdas(self, value): os.environ['OSBOT_LAMBDA_S3_FOLDER_LAMBDAS'  ] = value ; return value
    def set_lambda_role_name        (self, value): os.environ['OSBOT_LAMBDA_ROLE_NAME'          ] = value ; return value
    def set_bot_name                (self, value): os.environ['OSBOT_NAME'                      ] = value ; return value

    def sts__session_account_id(self):                   # to handle when the AWS_ACCOUNT_ID is not set
        #if self.aws_configured():                      # todo: find an efficient way to handle this since this doesn't work when a role is attached to a lambda function or EC2 instance
        from osbot_aws.aws.sts.STS import STS           #   the use of this method is not advised
        return STS().current_account_id()               #   since this is quite an expensive method

    def sts__caller_identity_user(self):
        #if self.aws_configured():                      # todo: find an efficient way to handle this since this doesn't work when a role is attached to a lambda function or EC2 instance
        from osbot_aws.aws.sts.STS import STS           #
        return STS().caller_identity_user()                  #

    # helper methods
    def account_id (self):
        return self.aws_session_account_id()

    def aws_configured(self):           # todo: add support for when the AWS is configured in EC2's account
        if self.aws_access_key_id():
            if self.aws_secret_access_key():
                if self.aws_session_region_name():
                    return True
        return False

    def lambda_s3_bucket(self):
        return self.resolve_lambda_bucket_name()

    def region_name(self):
        return self.aws_session_region_name()

    def resolve_lambda_bucket_name(self):
        bucket_name = os.getenv('OSBOT_LAMBDA_S3_BUCKET')
        if bucket_name is None:
            bucket_name = f'{self.aws_session_account_id()}--osbot-lambdas--{self.region_name()}' # this is a needed breaking change
        return bucket_name

    def resolve_temp_data_bucket_name(self):
        bucket_name = os.getenv('OSBOT_TEMP_DATA_S3_BUCKET')
        if bucket_name is None:
            bucket_name = f'{self.aws_session_account_id()}--temp-data--{self.region_name()}'
        return bucket_name

    def set_region(self,region_name):
        self.set_aws_session_region_name(region_name)

    def temp_data_bucket(self):
        return self.resolve_temp_data_bucket_name()

def set_aws_region(region_name):
    AWS_Config().set_aws_session_region_name(region_name)

aws_config = AWS_Config()


from osbot_aws.aws.s3.S3                    import S3
from osbot_utils.utils.Misc                 import random_string_and_numbers
from osbot_aws.AWS_Config                   import AWS_Config
from osbot_utils.testing.Temp_Zip_In_Memory import Temp_Zip_In_Memory


class S3_Zip_From_Files(Temp_Zip_In_Memory):

    def __init__(self, file_name=None, s3_bucket=None, s3_prefix=None):
        super().__init__()
        self.aws_config     = AWS_Config()
        self.s3             = S3()
        self.file_name      = file_name or f"temp_zip_file_{random_string_and_numbers()}"
        self.s3_bucket      = s3_bucket or self.aws_config.lambda_s3_bucket()
        self.s3_prefix      = s3_prefix or f'{AWS_Config().lambda_s3_folder_lambdas()}/unit_tests/temp_zips'
        self.s3_key         = f'{self.s3_prefix}/{self.file_name}.zip'

    def __enter__(self):
        super().__enter__()
        return self

    def __exit__(self, type, value, traceback):
        super().__exit__(type, value, traceback)
        pass

    def s3_file_create(self):
        kwargs = dict(file_body = self.zip_bytes(),
                      bucket    = self.s3_bucket  ,
                      key       = self.s3_key     )
        return self.s3.file_upload_from_bytes(**kwargs)

    def s3_file_delete(self):
        return self.s3.file_delete(self.s3_bucket, self.s3_key)

    def s3_file_exists(self):
        return self.s3.file_exists(self.s3_bucket, self.s3_key)

    def s3_file_contents(self):
        return self.s3.file_bytes(self.s3_bucket, self.s3_key)


from unittest import TestCase
from osbot_aws.aws.dynamo_db.Dynamo_DB import Dynamo_DB
from osbot_aws.aws.dynamo_db.Dynamo_DB__with_temp_role import Dynamo_DB__with_temp_role
from osbot_aws.testing.Pytest import skip_pytest___aws_pytest_user_name__is_not_set


class TestCase__Dynamo_DB(TestCase):
    dynamo_db      : Dynamo_DB
    reset_iam_creds: bool = False

    @classmethod
    def setUpClass(cls):
        skip_pytest___aws_pytest_user_name__is_not_set()
        cls.dynamo_db = Dynamo_DB__with_temp_role()
        if cls.reset_iam_creds:
            cls.dynamo_db.temp_role__iam_reset_credentials()


    @classmethod
    def tearDownClass(cls):
        pass


from os import environ

from dotenv                                                     import load_dotenv
from osbot_utils.decorators.lists.group_by                      import group_by
from osbot_utils.decorators.lists.index_by                      import index_by
from osbot_utils.helpers.sqlite.cache.Sqlite__Cache__Requests   import Sqlite__Cache__Requests
from osbot_utils.utils.Json                                     import json_loads
from osbot_utils.utils.Lists                                    import list_group_by
from osbot_utils.utils.Str                                      import str_dedent

ENV_NAME_BEDROCK_DB_NAME       = 'BEDROCK_DB_NAME'
SQLITE_DB_NAME__SQLITE_BEDROCK = 'sqlite__bedrock.sqlite'
SQLITE_TABLE__BEDROCK_REQUESTS = 'bedrock_requests'

class Bedrock__Cache(Sqlite__Cache__Requests):
    db_name        : str
    table_name     : str

    def __init__(self, db_path=None):
        load_dotenv()
        self.db_name    = environ.get(ENV_NAME_BEDROCK_DB_NAME) or SQLITE_DB_NAME__SQLITE_BEDROCK
        self.table_name = SQLITE_TABLE__BEDROCK_REQUESTS
        super().__init__(db_path=db_path, db_name=self.db_name, table_name=self.table_name)

    # Sqlite__Cache__Requests override methods

    def cache_request_data(self, model_id, body):
        request_data = dict(model        = model_id     ,
                            body         = body         )
        return request_data


    # Bedrock cache specific methods

    def comments(self, model):
        model_id = model.model_id
        body     = model.body()
        return self.cache_entry_comments(model_id, body)

    def comments_print(self, model):
        model_id = model.model_id
        body     = model.body()
        print()
        print()
        with_prompt = ""
        if hasattr(model, 'text_prompts'):
            with_prompt = "with prompt     : {model.text_prompts}\n"

        print(str_dedent(f"""
                ----------------------------
                comment on model: {model_id} "
                {with_prompt}            
                {self.cache_entry_comments(model_id, body)}"""))

    def comments_set(self, model, new_comments):
        model_id = model.model_id
        body     = model.body()
        return self.cache_entry_comments_update(model_id=model_id, body=body, new_comments=new_comments)

    def requests_data__all(self):                                               # this fixes the change in behaviour of requests_data__all which now returns the data in the request_data object
        requests_data__all = super().requests_data__all()
        for item in requests_data__all:
            request_data_raw      = item.get('request_data')                    # get the request_data value
            request_data_original = json_loads(request_data_raw)                # convert to object
            del item['request_data']                                            # delete the original entry to request_data
            item.update(request_data_original)                                  # add the request_data vars to the item of the array to return
        return requests_data__all

    def requests_data__by_model_id(self):
        values        = self.requests_data__all()
        group_by      = 'model'
        requests_data = list_group_by(values=values, group_by=group_by)
        if 'None' in requests_data:                         # if there are other entries in the cache
            del requests_data['None']                       # don't include them
        return requests_data

    def requests_data__with_model_id(self, model_id):
        return self.requests_data__by_model_id().get(model_id, [])


    # Bedrock overwrite methods

    def model_invoke(self, bedrock, model_id, body):
        target           = bedrock.model_invoke
        target_args      = []
        target_kwargs    = {'model_id': model_id, 'body': body}
        return self.invoke_with_cache(target, target_args, target_kwargs)


    def model_invoke_stream(self,bedrock, model_id, body):
        def capture_chunks():
            response = bedrock.model_invoke_stream(model_id, body)
            chunks   = []
            for chunk in response:
                chunks.append(chunk)
            return chunks

        target           = capture_chunks
        target_args      = []
        target_kwargs    = {}
        request_data     = dict(model_id=model_id, body=body, mode='invoke_stream')
        return self.invoke_with_cache(target, target_args, target_kwargs, request_data)


    @index_by
    @group_by
    def models(self, bedrock):                          # todo refactor out the caching login with this and model_invoke methods
        target           = bedrock.models
        target_args      = []
        target_kwargs    = {}
        request_data     = dict(method='models')
        return self.invoke_with_cache(target, target_args, target_kwargs, request_data)





# todo: refactor to osbot_dock since that is where the API_Docker and Docker_Image exist
from osbot_aws.AWS_Config import AWS_Config
from osbot_aws.apis.ECR import ECR

from osbot_utils.utils.Files import path_combine

class Create_Image_ECR:

    def __init__(self, image_name, path_images=None, image_tag='latest') -> object:
        from osbot_docker.apis.API_Docker import API_Docker
        from osbot_docker.apis.Docker_Image import Docker_Image
        self.api_docker      = API_Docker()
        self.ecr             = ECR()
        self.aws_config      = AWS_Config()
        self.image_name      = image_name
        self.image_tag       = image_tag
        self.path_images     = path_images or path_combine(__file__, '../../images')
        self.docker_image    = Docker_Image(image_name=self.image_repository(),image_tag=self.image_tag, api_docker=self.api_docker)

    def build_image(self):
        result       = self.docker_image.build(self.path_image())
        return result

    def create_repository(self):
        self.ecr.repository_create(self.image_name)
        return self.repository_exists()

    def delete_repository(self):
        return self.ecr.repository_delete(self.image_name)

    def full_image_name(self):
        return f'{self.image_repository()}:{self.image_tag}'

    def image_repository(self):
        account_id = self.aws_config.aws_session_account_id()
        region     = self.aws_config.aws_session_region_name()
        return f'{account_id}.dkr.ecr.{region}.amazonaws.com/{self.image_name}'

    def ecr_login(self):
        auth_data = self.ecr.authorization_token()
        return self.api_docker.registry_login(registry=auth_data.get('registry'),
                                              username=auth_data.get('username'),
                                              password=auth_data.get('password'))


    def path_image(self):
        return path_combine(self.path_images, self.image_name)

    def push_image(self):
        auth_result = self.ecr_login()                             # make sure we are logged in
        push_json_lines = self.docker_image.image_push()
        return dict(auth_result=auth_result, push_json_lines=push_json_lines)

    def run_locally(self):
        full_image_name = self.full_image_name()
        return self.api_docker.docker_run(full_image_name)

    def create(self):
        create_repository = self.create_repository   ()
        ecr_login         = self.ecr_login           ()
        build_image       = self.build_image         ()
        push_image        = self.push_image          ()
        return {'create_repository' : create_repository,
                'ecr_login'         : ecr_login        ,
                'build_image'       : build_image      ,
                'push_image'        : push_image       }

    def repository_exists(self):
        return self.ecr.repository_exists(self.image_name)



from osbot_aws.aws.s3.S3           import S3
from osbot_utils.helpers.Zip_Bytes import Zip_Bytes


class S3__Zip_Bytes(Zip_Bytes):
    s3: S3

    def save_to_s3(self, s3_bucket, s3_key):
        kwargs = dict(file_bytes = self.zip_bytes,
                      bucket     = s3_bucket     ,
                      key        = s3_key        )
        return self.s3.file_create_from_bytes(**kwargs)



from os import chmod

from osbot_aws.aws.ec2.AMI import AMI
from osbot_aws.aws.ec2.EC2                          import EC2
from osbot_utils.base_classes.Type_Safe             import Type_Safe
from osbot_utils.decorators.methods.cache_on_self   import cache_on_self
from osbot_utils.utils.Files import current_temp_folder, folder_create, parent_folder, file_exists, file_create, \
    file_not_exists

DEFAULT__EC2__SECURITY_GROUP_NAME__WITH_SSH = 'security_group_with_only_ssh'
DEFAULT__EC2__KEY_NAME_FORMAT               = '{account_id}__osbot__{region_name}'
DEFAULT__EC2__KEY_FILE__LOCATION            = '{current_temp_folder}/osbot_aws_keys/{key_name}.pem'


# todo:
# need to make sure that the current identity has ec2:* and these extra iam permissions
# {
#     "Version": "2012-10-17",
#     "Statement": [
#         {
#             "Sid": "VisualEditor0",
#             "Effect": "Allow",
#             "Action": [
#                 "iam:CreateServiceLinkedRole",
#                 "iam:PutRolePolicy"
#             ],
#             "Resource": "*"
#         }
#     ]
# }
# this should be in a generic policy called something like Policy__EC__required_to_create_instances
class EC2__Create__Instance(Type_Safe):
    image_id          : str
    instance_type     : str
    key_name          : str
    security_group_id : str
    spot_instance     : bool = True
    tags              : dict
    ec2               : EC2

    def create_kwargs(self):
        return  dict(image_id          = self.image_id           ,
                     instance_type     = self.instance_type      ,
                     key_name          = self.key_name           ,
                     security_group_id = self.security_group_id  ,
                     spot_instance     = self.spot_instance      ,
                     tags              = self.tags               )

    def ami_amazon_linux_3_x86_64(self):
        return AMI().amazon_linux_3_x86_64()


    def create_instance(self):
        create_kwargs = self.create_kwargs()
        return self.ec2.instance_create(**create_kwargs)

    def create_instance__with_amazon_linux__t3_nano__ssh_sg__spot(self):
        self.setup__with_amazon_linux__t3_nano__ssh_sg__spot()
        return self.create_instance()

    def setup__with_amazon_linux__t3_nano__ssh_sg__spot(self):
        self.image_id          = self.ami_amazon_linux_3_x86_64()
        self.instance_type     = 't3.nano'
        self.key_name          = self.key_name__create_if_doesnt_exist()
        self.security_group_id = self.security_group_with_ssh()
        self.spot_instance     = True
        self.tags              = {'created-by': 'OSBot_AWS.Create_EC2_Instance'}


    def key_name__create_if_doesnt_exist(self):
        key_name = self.key_name__from_account_id_and_region_name()
        key_path = self.path_key_file()
        key_details = self.ec2.key_pair(key_pair_name=key_name)
        if key_details:
            if file_not_exists(key_path):   # f"Key pair {key_name} exists in EC2 but the local key file not found: {key_path}")
                self.ec2.key_pair_delete(key_pair_name=key_name)
            else:
                return key_name

        key_data     = self.ec2.key_pair_create(key_name=key_name)
        key_contents = key_data.get('KeyMaterial')
        file_create(key_path, key_contents)             # create file
        chmod(path=key_path, mode= 0o400)               # set the correct permissions
        return key_name

    def key_name__from_account_id_and_region_name(self):
        key_name = DEFAULT__EC2__KEY_NAME_FORMAT.format(account_id=self.ec2.aws_config.account_id(), region_name=self.ec2.aws_config.region_name())
        return key_name

    def path_key_file(self):
        key_name = self.key_name__from_account_id_and_region_name()
        key_file = DEFAULT__EC2__KEY_FILE__LOCATION.format(current_temp_folder=current_temp_folder(), key_name=key_name)
        folder_create(parent_folder(key_file))
        return key_file

    def security_group_with_ssh(self):
        security_group = self.ec2.security_group(security_group_name=DEFAULT__EC2__SECURITY_GROUP_NAME__WITH_SSH)
        if security_group is None:
            # create security group
            security_group_name = DEFAULT__EC2__SECURITY_GROUP_NAME__WITH_SSH
            description         = DEFAULT__EC2__SECURITY_GROUP_NAME__WITH_SSH
            result              = self.ec2.security_group_create(security_group_name=security_group_name, description=description)
            security_group_id   = result.get('data').get('security_group_id')
            # enable SSH
            self.ec2.security_group_authorize_ingress(security_group_id, port=22)
            return security_group_id

        security_group_id = security_group.get('GroupId')
        return security_group_id








from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Png import save_png_base64_to_file

# see https://docs.aws.amazon.com/bedrock/latest/userguide/titan-image-models.html
#     https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters-titan-image.html
class Amazon_Titan_Image_Generator_V1(Kwargs_To_Self):
    model_id         : str = 'amazon.titan-image-generator-v1'
    text             : str
    negative_text    : str
    number_of_images : int  = 1
    height           : int  = 512  #512
    width            : int  = 512  #512
    cfg_scale        : float = 10.0
    seed             : int   = 42           # just because

    # IMAGE_VARIATION
    def body__image_variation(self, images):
        image_variation_params = dict(**self.text_and_negative_text(),
                                      images=images)
        return {"taskType"             : "IMAGE_VARIATION"              ,
                "imageVariationParams" : image_variation_params         ,
                "imageGenerationConfig": self.image_generation_config() }

    # INPAINTING
    def body__in_painting(self, image, mask_prompt):
        in_painting_params = dict(**self.text_and_negative_text(),
                                  image      = image             ,
                                  maskPrompt = mask_prompt       )
        return {"taskType"             : "INPAINTING",
                "inPaintingParams"     : in_painting_params,
                "imageGenerationConfig": self.image_generation_config()}

    # INPAINTING
    def body__out_painting(self, image, mask_prompt):
        out_painting_params = dict(**self.text_and_negative_text(),
                                  image           = image             ,
                                  maskPrompt      = mask_prompt       ,
                                  outPaintingMode = "DEFAULT"         )
        return {"taskType"             : "OUTPAINTING",
                "outPaintingParams"    : out_painting_params,
                "imageGenerationConfig": self.image_generation_config()}

    #TEXT_IMAGE

    def body__text_image(self):
        text_to_image_params = { "text": self.text }
        if self.negative_text and len(self.negative_text) > 3:                  # or it will fail AWS validation
            text_to_image_params["negativeText"] = self.negative_text
        return {"taskType"             : "TEXT_IMAGE"                             ,
                "textToImageParams"    : text_to_image_params                     ,
                "imageGenerationConfig": { "numberOfImages": self.number_of_images,
                                           "height"        : self.height          ,
                                           "width"         : self.width           ,
                                           "cfgScale"      : self.cfg_scale       ,
                                           "seed"          : self.seed            }}



    def image_generation_config(self):
        return { "numberOfImages": self.number_of_images,
                 "height"        : self.height          ,
                 "width"         : self.width           ,
                 "cfgScale"      : self.cfg_scale       }

    def text_and_negative_text(self):
        result = {"text": self.text}
        if self.negative_text and len(self.negative_text) > 3:  # or it will fail AWS validation
            result["negativeText"] = self.negative_text
        return result

    def save_png_data(self, png_data, target_file=None):
        if target_file is None:
            target_file = "/tmp/bedrock_image.png"
        save_png_base64_to_file(png_data, target_file)


from osbot_aws.aws.apigateway.lambdas.lambda_web_sockets import run
from osbot_aws.deploy.Deploy_Lambda import Deploy_Lambda


class Deploy__lambda_web_sockets:

    def __init__(self):
        self.handler       = run
        self.deploy_lambda = Deploy_Lambda(self.handler)

    def update(self):
        self.deploy_lambda.add_osbot_utils()
        return self.deploy_lambda.update()

    def invoke_lambda(self, payload=None):
        return self.deploy_lambda.invoke(payload)



from os import environ

from osbot_aws.aws.ec2.AMI import AMI
from osbot_aws.aws.ec2.EC2                      import EC2
from osbot_utils.base_classes.Kwargs_To_Self    import Kwargs_To_Self
from osbot_utils.helpers.ssh.SSH                import SSH
from osbot_utils.helpers.ssh.SSH__Execute       import ENV_VAR__SSH__HOST, ENV_VAR__SSH__KEY_FILE, ENV_VAR__SSH__USER
from osbot_utils.utils.Dev                      import pprint
from osbot_utils.utils.Http                     import wait_for_ssh
from osbot_utils.utils.Misc                     import random_string

class EC2_Instance(Kwargs_To_Self):
    instance_id   : str
    create_kwargs : dict
    ec2           : EC2
    # def __init__(self, instance_id=None, create_kwargs=None):
    #     self.create_kwargs = create_kwargs
    #     self.instance_id   = instance_id
    #     self.ec2           = EC2()

    def create(self, instance_name = None):
        kwargs = self.create_kwargs
        random_name = random_string(prefix='test_ec2_with_ssh_support')
        if not kwargs.get('image_id'     ): kwargs['image_id'     ] = AMI().amazon_linux_3()
        if not kwargs.get('instance_type'): kwargs['instance_type'] = 't2.nano'
        if not kwargs.get('name'         ): kwargs['name'         ] = instance_name or random_name
        if not kwargs.get('tags'         ): kwargs['tags'         ] = {'Type': 'EC2_Instance'}

        self.instance_id = self.ec2.instance_create(**kwargs)
        return self.instance_id

    def delete(self):
        return self.ec2.instance_delete(instance_id=self.instance_id)

    def exists(self):
        return self.ec2.instance_exists(instance_id=self.instance_id)

    def info(self):
        return self.ec2.instance_details(instance_id=self.instance_id)

    def print_info(self):
        pprint(self.info())

    def ip_address(self):
        return self.info().get('public_ip')

    def start(self):
        return self.ec2.instance_start(instance_id=self.instance_id)

    def stop(self):
        return self.ec2.instance_terminate(instance_id=self.instance_id)

    def ssh(self, ssh_key_file, ssh_key_user):
        ip_address = self.ip_address()
        ssh = SSH()
        environ[ENV_VAR__SSH__HOST    ] = ip_address
        environ[ENV_VAR__SSH__KEY_FILE] = ssh_key_file          # todo: redo this workflow on how to set up the ssh_execute obejct
        environ[ENV_VAR__SSH__USER    ] = ssh_key_user
        ssh.ssh_execute().setup()
        return ssh

    def state(self):
        return self.info().get('state')

    def tags(self):
        return self.info().get('tags')

    def wait_for_ssh(self):
        return wait_for_ssh(self.ip_address())


from datetime import datetime, timezone

import boto3
from botocore.exceptions import ClientError

from osbot_utils.decorators.methods.cache import cache
from osbot_utils.helpers.Local_Cache import Local_Cache
from osbot_aws.aws.iam.IAM import IAM
from osbot_aws.aws.iam.IAM_Role import IAM_Role
from osbot_aws.aws.sts.STS import STS
from osbot_utils.utils.Misc import wait_for
from osbot_utils.utils.Str import safe_str

CACHES_NAME__AWS_ROLES       = '_aws_cached_roles'

#FOLDER_NAME__CACHED_CREDENTIALS = '_aws_cached_credentials'


class IAM_Assume_Role:

    def __init__(self, role_name, policies_to_add=None):
        self.policies_to_add  = policies_to_add or []
        self.role_name        = role_name
        self.cached_role      = Local_Cache(role_name, CACHES_NAME__AWS_ROLES)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    @cache
    def iam(self) -> IAM:
        return IAM()

    @cache
    def sts(self) -> STS:
        return STS()
    @cache
    def iam(self):
        return self.iam_role().iam

    @cache
    def iam_role(self) -> IAM_Role:
        return IAM_Role(role_name=self.role_name)

    def aws__client__sts(self):
        class STS__Assume_Role(STS):
            def client(_self):
                return self.boto3_client(service_name='sts')
        return STS__Assume_Role()

    def add_policy(self, service, action, resource):
        if service and action and resource:
            policy_name = f'inline_policy_for_{safe_str(service)}_{safe_str(action)}_{safe_str(resource)}'
            policy_document = self.create_policy_document(service=service, action=action, resource=resource)
            return self.set_inline_policy(policy_name, policy_document)
        return False
    def assume_policy(self):
        return self.setup_data().get('assume_policy')

    def boto3_client(self, service_name='iam', region_name=None, retries=20):
        credentials = self.credentials(retries=retries)
        kwargs = dict(service_name          = service_name                  ,
                      aws_access_key_id     = credentials['AccessKeyId'    ],
                      aws_secret_access_key = credentials['SecretAccessKey'],
                      aws_session_token     = credentials['SessionToken'   ])
        if region_name:
            kwargs['region_name'] = region_name
        return boto3.client(**kwargs)

    def boto3_resource(self, service_name='iam', region_name=None, retries=20):
        credentials = self.credentials(retries=retries)
        kwargs = dict(service_name          = service_name                  ,
                      aws_access_key_id     = credentials['AccessKeyId'    ],
                      aws_secret_access_key = credentials['SecretAccessKey'],
                      aws_session_token     = credentials['SessionToken'   ])
        if region_name:
            kwargs['region_name'] = region_name
        return boto3.resource(**kwargs)

    def create_credentials(self, retries=30):
        self.credentials_raw(retries=retries)
        return self

    def create_policy_document(self, service, action, resource, effect='Allow'):
        return { "Version"  : "2012-10-17",
                 "Statement": [{"Effect" : effect               ,
                               "Action"  : f"{service}:{action}",
                               "Resource": resource             }]}

    def create_policies(self):
        for policy_to_add in self.policies_to_add:
            self.add_policy(**policy_to_add)

    def create_role(self, recreate=False, create_credentials=True):
        if recreate:
            self.reset()
            self.delete_role()

        if self.role_exists() is False:
            self.setup_data()
            assume_policy_document = self.assume_policy()
            result__role_create    = self.iam_role().create(assume_policy_document=assume_policy_document)
            self.cached_role.set('result__role_create', result__role_create)
            self.cached_role.set('role_exists', True                 )  # todo see how to check this would having to make another Boto3 call
            self.create_policies()
            if create_credentials:
                self.create_credentials()

    def credentials(self, reset=False, retries=30):
        if reset or self.credentials_expired():
            self.credentials_reset()

        credentials_raw = self.credentials_raw(retries=retries)
        if credentials_raw:
            credentials_data = credentials_raw.get('Credentials', {})
            return { 'AccessKeyId'    : credentials_data.get('AccessKeyId'    ),
                     'SecretAccessKey': credentials_data.get('SecretAccessKey'),
                     'SessionToken'   : credentials_data.get('SessionToken'   )}
        return {}

    def credentials_expiration_date(self):
        credentials_raw = self.data().get('result__credentials') or {}
        expiration  = credentials_raw.get('Credentials', {}).get('Expiration')
        if expiration:
            if type(expiration) is datetime:
                return expiration
            if type(expiration) is str:
                return  datetime.strptime(expiration, '%Y-%m-%d %H:%M:%S%z')
        return None

    def credentials_expired(self):
        expiration_time = self.credentials_expiration_date()
        if expiration_time:
            current_time = datetime.now(timezone.utc)
            return expiration_time < current_time
        return True

    def credentials_raw(self, role_session_name=None,  retries=30, delay=0.5):      # todo: capture metric for how long it takes to get the credentials
        credentials_raw = self.data().get('result__credentials')
        if credentials_raw is not None:
            return credentials_raw
        for attempt in range(retries):
            try:
                credentials_raw = self.sts().assume_role(role_arn=self.role_arn(), role_session_name=role_session_name)
                self.cached_role.set('result__credentials', credentials_raw)
                return credentials_raw
            except ClientError as e:
                if e.response['Error']['Code'] == 'AccessDenied' and attempt < retries - 1:
                    wait_for(delay)
                else:
                    print(">>>>>>>> in credentials_raw <<<<<< FAILED TO GET CREDENTIALS")
                    raise
        raise Exception(">>>>>>>> in credentials_raw <<<<<< FAILED TO GET CREDENTIALS")


    def credentials_reset(self):
        self.cached_role.set('result__credentials', None)
        return self

    def data(self):
        return self.cached_role.data()

    def delete_role(self):
        self.iam_role().delete()
        self.cached_role.set('role_exists', False)             # todo see how to check this would having to make another Boto3 call
        return self


    def default_assume_policy(self, user_arn=None, service_name=None, federated=None, canonical_user=None):
        effect     = 'Allow'
        action     = 'sts:AssumeRole'
        statements = []
        def add_statement(name, target):
            if target:
                statements.append({ 'Effect': effect, 'Principal': { name: target }, 'Action': action })

        add_statement('AWS'          , user_arn      )
        add_statement('Service'      , service_name  )
        add_statement('Federated'    , federated     )
        add_statement('CanonicalUser', canonical_user)

        return { 'Version'  : '2012-10-17',
                 'Statement': statements  }

    def policies(self):
        policies = self.data().get('policies')
        if policies is None:
            policies = self.iam_role().iam.role_policies_inline(fetch_inline_policies=True)
            self.cached_role.set('policies', policies)
        return policies

    def role_arn(self):
        return self.data().get('role_arn')

    def role_exists(self):
        return self.cached_role.get('role_exists', False)


    def set_inline_policy(self, policy_name, policy_document):
        policies = self.policies()
        if (policy_name in policies) is False or policies.get(policy_name) != policy_document:
            self.iam_role().iam.role_policy_add(policy_name, policy_document)
            policies[policy_name] = policy_document
            self.cached_role.set('policies', policies)
            return True
        return False

    def setup_data(self):
        if self.cached_role.cache_exists() is False:
            role_name          = self.role_name
            caller_identity    = self.sts().caller_identity()
            current_user_arn   = caller_identity.get('Arn')
            assume_policy      = self.default_assume_policy(user_arn=current_user_arn)
            current_account_id = caller_identity.get('Account')
            current_user_id    = caller_identity.get('UserId' )
            policies_to_add    = self.policies_to_add

            role_arn           = f"arn:aws:iam::{current_account_id}:role/{role_name}"

            data = dict(assume_policy       = assume_policy                 ,
                        current_account_id  = current_account_id            ,
                        current_user_id     = current_user_id               ,
                        current_user_arn    = current_user_arn              ,
                        policies            = None                          ,           # setting these vars to None will trigger the reload on the first time they are used
                        policies_to_add     = policies_to_add               ,
                        result__credentials = None                          ,
                        result__role_create = None                          ,
                        role_arn            = role_arn                      ,
                        role_name           = role_name                     ,
                        role_exists         = self.iam_role().exists()      )
            self.cached_role.set_data(data)
        return self.cached_role.data()

    def reset(self):
        self.cached_role.cache_delete()
        self.setup_data()
        return self

    def wait_for_valid_execution(self, service_name, method_name, retries=20, delay=0.2):
        for attempt in range(retries):
            try:
                target_client = self.boto3_client(service_name)
                target_method = getattr(target_client, method_name)
                return target_method()
            except Exception as e:
                print()
                print(f">> [ attempt {attempt}] in wait_for_valid_execution | waiting for {delay} seconds before retrying << ")
                print(f">>   error:{e}")
            wait_for(delay)
            self.credentials_reset()                # reset credentials before trying again


from osbot_aws.apis.Logs import Logs


def run(event, context):
    group_name   = event.get('group_name' )
    stream_name  = event.get('stream_name')
    message      = event.get('message'    )
    logs         = Logs(group_name=group_name,
                        stream_name=stream_name)
    return logs.event_add(message)


from osbot_aws.Dependencies import Dependencies

from osbot_utils.utils.Files import path_combine, folder_exists


# todo: refactor this entire workflow to use python wheels which seem a much better solution
class Lambda_Upload_Package:

    def __init__(self):
        self.path_packages = path_combine('.', '../../../_lambda_dependencies/')
        self.dependencies  = Dependencies()

    def local_path(self, package):
        return path_combine(self.path_packages, package)

    def exists_locally(self, package):
        return folder_exists(self.local_path(package))

    def install_locally(self, package):
        if self.exists_locally(package) is False:
            self.dependencies.pip_install_dependency(package)
        return self.exists_locally(package)

    def upload_to_s3(self, package):
        if type(package) is list:
            for item in package:
                self.upload_to_s3(item)
        else:
            if self.install_locally(package):
                return self.dependencies.upload_dependency(package)

    # def test_install_locally(self):
    #     assert self.lambda_setup.install_locally(self.package) is True
    #     assert self.lambda_setup.exists_locally (self.package) is True
    #
    # def test_upload_to_s3(self):
    #     pprint(self.lambda_setup.    upload_to_s3(self.package))



from osbot_aws.aws.dynamo_db.Dynamo_DB__Table       import Dynamo_DB__Table
from osbot_utils.utils.Misc                         import list_set


class DyDB__Table(Dynamo_DB__Table):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def add_document(self, document):
        result = super().add_document(document)
        if result.get('status') == 'ok':
            return result.get('data')
        return result

    def add_documents(self,documents):
        return super().add_documents(documents).get('data')

    def create_table_kwargs(self):
        key_schema               = [{'AttributeName': self.key_name  , 'KeyType'      : 'HASH'}]
        attribute_definitions    = [{'AttributeName': self.key_name  , 'AttributeType': 'S'   }]
        #global_secondary_indexes = []

        # for index in self.global_secondary_indexes:
        #     attribute_definitions   .append({ 'AttributeName': index, 'AttributeType': 'S' })
        #     global_secondary_indexes.append({ 'IndexName': f'{index}_index',
        #                                       'KeySchema': [{'AttributeName': index, 'KeyType': 'HASH'}],
        #                                       'Projection': {'ProjectionType': 'ALL'} })
        kwargs   = { 'AttributeDefinitions'  : attribute_definitions    ,
                     'BillingMode'           : 'PAY_PER_REQUEST'        ,
                     'KeySchema'             : key_schema               ,
                     #'GlobalSecondaryIndexes': global_secondary_indexes ,
                     'TableName'             : self.table_name           }
        return kwargs

    def create_table(self, wait_for_table=True):
        if not self.table_name:
            raise ValueError('Table name is required')
        if self.exists():
            return False

        kwargs = self.create_table_kwargs()
        self.dynamo_db.client().create_table(**kwargs)
        if wait_for_table:
            self.dynamo_db.wait_for('table_exists', self.table_name)
        return True

    def delete_document(self, document_id):
        result = super().delete_document(key_value=document_id)
        return result.get('status') == 'ok'

    def delete_documents(self, documents_ids):
        return super().delete_documents(keys_values=documents_ids).get('data')

    def delete_table(self, wait_for_deletion=False):
        return super().delete_table(wait_for_deletion=wait_for_deletion).get('data')

    def document(self, document_id):
        return super().document(key_value=document_id).get('data')

    def document_exists(self, document_id):
        return self.document(document_id) != {}

    def documents(self, documents_ids):
        return super().documents(keys_values=documents_ids).get('data').get('all_responses')

    def documents_all(self, **kwargs):
        return super().documents_all().get('data')

    def documents_ids(self, **kwargs):
        return super().documents_ids(**kwargs).get('data')

    def dydb_document(self, document_id, load_data=True):
        from osbot_aws.aws.dynamo_db.models.DyDB__Document import DyDB__Document        # need to import this here due to circular references
        document = None                                                                 # create var to hold document data
        if load_data:                                                                   # if asked to load data
            document = self.document(document_id)                                       # try to load data (this will return None if the document doesn't exist)
        if not document:                                                                # if load_data is False or self.document(document_id) returned none
            document = {self.key_name: document_id}                                     # create a default document representation with only the id
        kwargs_document = dict(document = document,                                     # kwargs for DyDB__Document object
                               table    = self    )                                     #     where pass a reference to the current table
        return DyDB__Document(**kwargs_document)                                        # create DyDB__Document

    def exists(self):
        return super().exists().get('data')

    def info(self):
        return super().info().get('data')

    def not_exists(self):
        return self.exists() is False

    def size(self):
        return self.dynamo_db.documents_count(self.table_name)

    def query_index(self, index_name, index_type, index_value):
            query_kwargs = dict( TableName                 = self.table_name                                ,
                                 IndexName                 = index_name                                     ,
                                 KeyConditionExpression    =  f'#{index_name} = :{index_name}'              ,
                                 ExpressionAttributeNames  = { f'#{index_name}' : index_name                },
                                 ExpressionAttributeValues = { f':{index_name}' : {index_type :index_value  }})
            response = self.query(**query_kwargs)

            items = response.get('data', {}).get('Items', [])           # todo: add cases with large results
            return [self.dynamo_db.document_deserialize(item) for item in items]

    def query_index_between_range(self, index_name, index_type, index_value, sort_key, sort_key_type, start_value, end_value, query_filter=None):
        query_kwargs = dict( TableName                 = self.table_name                        ,
                             IndexName                 = index_name                 ,
                             KeyConditionExpression    =  f'#{index_name} = :{index_name} AND #{sort_key} BETWEEN :start AND :end',
                             ExpressionAttributeNames  = { f'#{index_name}' : index_name,   f'#{sort_key}': sort_key }            ,
                             ExpressionAttributeValues = { f':{index_name}' : {index_type :index_value    },
                                                           ':start'         : { sort_key_type : str(start_value)},
                                                           ':end'           : { sort_key_type : str(end_value  )}})
        if query_filter and list_set(query_filter) == ['expression_attr_names', 'expression_attr_values', 'filter_expression']:
            query_kwargs['ExpressionAttributeNames' ].update(query_filter.get('expression_attr_names'))
            query_kwargs['ExpressionAttributeValues'].update(query_filter.get('expression_attr_values'))
            query_kwargs['FilterExpression'         ] = query_filter.get('filter_expression')

        #pprint(query_kwargs)
        response = self.query(**query_kwargs)

        items = response.get('data', {}).get('Items', [])           # todo: add cases with large results
        return [self.dynamo_db.document_deserialize(item) for item in items]


from osbot_aws.apis.Events import Events
from osbot_aws.helpers.Event_Rule import Event_Rule
from osbot_utils.utils.Misc import random_string


class Temp_Event_Rule:

    def __init__(self, rule_name=None, event_source=None, tags=None, event_bus_name=None):
        self.event_rule = Event_Rule(rule_name=rule_name, event_source=event_source, tags=tags, event_bus_name=event_bus_name)

    def __enter__(self):
        self.event_rule.create()
        return self.event_rule


    def __exit__(self, type, value, traceback):
        self.event_rule.delete()



import datetime
import json

from osbot_utils.decorators.lists.group_by  import group_by
from osbot_utils.decorators.lists.index_by  import index_by
from osbot_utils.utils.Json                 import json_parse
from osbot_aws.apis.Boto_Helpers            import Boto_Helpers
from osbot_aws.aws.s3.S3                    import S3
from osbot_aws.apis.Session                 import Session


class Cloud_Trail(Boto_Helpers):                # todo write tests for this class (which was lost in a past refactor)

    def __init__(self):
        self.cloudtrail = Session().client('cloudtrail')
        self.s3         = S3()

    # helpers
    def date_now(self):
        return datetime.datetime.utcnow()

    def date_minutes_ago(self, minutes):
        return datetime.datetime.utcnow() - datetime.timedelta(minutes=minutes)

    def event_selectors(self, trail_name):
        return  self.cloudtrail.get_event_selectors(TrailName=trail_name).get('EventSelectors')

    def events(self, start_time, end_time, max_results=50, lookup_attributes=None):
        kwargs = {
            'LookupAttributes': lookup_attributes or [],
            'StartTime'       : start_time              ,
            'EndTime'         : end_time                ,
            'MaxResults'      : max_results
        }
        return self.cloudtrail.lookup_events(**kwargs).get('Events')

    def events_all(self, start_time, end_time, lookup_attributes=None, page_size=None):
        kwargs = {  'api'             : self.cloudtrail         ,
                    'method'          : 'lookup_events'         ,
                    'field_id'        : 'Events'                ,
                    'LookupAttributes': lookup_attributes or [] ,
                    'StartTime'       : start_time              ,
                    'EndTime'         : end_time                }
        if page_size:
            kwargs['MaxResults'] = page_size
        return Boto_Helpers.invoke_using_paginator(**kwargs)

    def events_in_last(self,minutes, lookup_attributes=None, page_size=None):
        end_time   = self.date_now()
        start_time = self.date_minutes_ago(minutes)
        return self.events_all(start_time, end_time, lookup_attributes=lookup_attributes, page_size=page_size)

    @index_by
    @group_by
    def convert_to_cloud_trail_events(self, events):
        for event in events:
            cloud_trail_event = json_parse(event.get('CloudTrailEvent', {}))
            yield cloud_trail_event
    def tags(self, resource_ids=[]):
        return self.cloudtrail.list_tags(ResourceIdList=resource_ids)

    def log_files(self, trail_name, account_id, region, year, month, day):
        trail         = self.trail(trail_name)
        log_type      = 'CloudTrail'
        s3_bucket     = trail.get('S3BucketName')
        s3_key_prefix = trail.get('S3KeyPrefix')
        s3_prefix     = f'{s3_key_prefix}/AWSLogs/{account_id}/'                           \
                        f'{log_type}/{region}/{year}/{month}/{day}/'
        s3_files      = self.s3.find_files(s3_bucket, prefix=s3_prefix)
        return s3_files

    def log_files_records(self, trail_name, log_files):
        trail     = self.trail(trail_name)
        s3_bucket = trail.get('S3BucketName')
        records   = []
        for log_file in log_files:
            contents = self.s3.file_contents_from_gzip(s3_bucket, log_file)
            records.extend(json.loads(contents).get('Records'))
        return records

    def trail(self, trail_name):
        trail        = self.cloudtrail.get_trail(Name=trail_name).get('Trail')
        trail_status = self.cloudtrail.get_trail_status(Name=trail_name)        # also get the trail status
        trail.update(trail_status)                                              # add its values to the trail object
        del trail['ResponseMetadata']                                           # remove the ResponseMetadata since we don't need this value
        return trail

    def trail_create(self, trail_name, account_id, region, s3_bucket, s3_key_prefix):
        self.trail_s3_bucket_configure(trail_name, s3_bucket, account_id, region)       # create and configure bucket
        self.trail_create_raw(trail_name, s3_bucket, s3_key_prefix)                                 # create trail
        self.trail_start(trail_name)                                                                # enable it
        return self.trail(trail_name)                                                               # return trail details

    def trail_create_raw(self, trail_name, s3_bucket, s3_key_prefix, sns_topic_name=None, global_service_events=True, multi_region_trail=True, log_file_verification=True):
        params = {  'Name'                      : trail_name,
                    'S3BucketName'              : s3_bucket,
                    'S3KeyPrefix'               : s3_key_prefix,
                    'IncludeGlobalServiceEvents': global_service_events,
                    'IsMultiRegionTrail'        : multi_region_trail,
                    'EnableLogFileValidation'   : log_file_verification
        }
        if sns_topic_name:
            params['SnsTopicName'] = sns_topic_name
        try:
            return self.cloudtrail.create_trail(**params)
        except Exception as error:
            return { 'error': f'{error}'}

    def trail_delete(self, trail_name):
        self.cloudtrail.delete_trail(Name=trail_name)
        return self

    def trail_status(self,trail_name):
        return self.cloudtrail.get_trail_status(Name=trail_name)

    def trail_s3_bucket_configure(self, trail_name, s3_bucket, account_id, region):

        self.s3.bucket_create(s3_bucket, region)                # make sure bucket exists
                                                                # see https://docs.aws.amazon.com/awscloudtrail/latest/userguide/create-s3-bucket-policy-for-cloudtrail.html
        resource_arn = self.s3.policy_statements__resource_arn(s3_bucket, trail_name, account_id)
        statements   = self.s3.policy_statements__without(s3_bucket, 'Resource', resource_arn) # making sure we are not adding the same resource_arn twice
        statement    = self.s3.policy_statements__new('s3:PutObject', 'Allow', 'cloudtrail.amazonaws.com', resource_arn)
        statements.append(statement)
        if 's3:GetBucketAcl' not in self.s3.policy_statements(s3_bucket, index_by='Action'):
            statements.append(self.s3.policy_statements__new('s3:GetBucketAcl', 'Allow', 'cloudtrail.amazonaws.com', f"arn:aws:s3:::{s3_bucket}"))

        self.s3.policy_create(s3_bucket,statements)             # add cloud trail policy
        return self

    def trail_start(self, trail_name):
        self.cloudtrail.start_logging(Name=trail_name)
        return self

    def trail_stop(self, trail_name):
        return self.cloudtrail.stop_logging(Name=trail_name)

    def trails(self):
        return self.cloudtrail.list_trails().get('Trails')



from osbot_aws.apis.CodeBuild import CodeBuild
from osbot_aws.aws.iam.IAM import IAM


#todo: add Create_Code_Build tests from gs-docker-codebuild project
class Create_Code_Build:

    def __init__(self, project_name, github_org='owasp-sbot', source_version='master', build_spec='buildspec.yml', docker_type='LINUX_CONTAINER',docker_image='aws/codebuild/docker:18.09.0',compute_type='BUILD_GENERAL1_SMALL'):
        self.iam                = IAM()
        self.build_spec         = build_spec
        self.github_org         = github_org
        self.source_version     = source_version
        self.docker_type        = docker_type
        self.docker_image       = docker_image
        self.compute_type       = compute_type
        self.project_name       = project_name
        self.account_id         = self.iam.account_id()
        self.region             = self.iam.region()
        self.project_repo       = f'https://github.com/{self.github_org}/{self.project_name}'
        self.service_role       = f'arn:aws:iam::{self.account_id}:role/{self.project_name}'
        self.project_arn        = f'arn:aws:codebuild:{self.region}:{self.account_id}:project/{self.project_name}'
        self.assume_policy      = {'Statement': [{'Action'   : 'sts:AssumeRole'                     ,
                                                  'Effect'   : 'Allow'                              ,
                                                  'Principal': {'Service': 'codebuild.amazonaws.com'}}]}

        self.code_build = CodeBuild(role_name=self.project_name, project_name=self.project_name)

    def create_role_and_policies(self, policies):
        self.code_build.iam.role_create(self.assume_policy)
        policies_arn = self.code_build.policies_create(policies)
        self.code_build.iam.role_policies_attach(policies_arn)
        return self

    def create_project_with_container__docker(self):
        kvargs = {
            'name'         : self.project_name,
            'source'       : { 'type'         : 'GITHUB',
                               'location'     : self.project_repo                  ,
                               'buildspec'    : self.build_spec                   },
            'sourceVersion': self.source_version                                   ,
            'artifacts'    : {'type'          : 'NO_ARTIFACTS'                    },
            'environment'  : {'type'          : self.docker_type                   ,
                              'image'         : self.docker_image                  ,
                              'computeType'   : self.compute_type                  },
            'serviceRole'  : self.service_role
        }

        return self.code_build.codebuild.create_project(**kvargs)

    def create_project_with_container__gs_docker_codebuild(self):
        kvargs = {
            'name'         : self.project_name,
            'source'       : { 'type'                   : 'GITHUB',
                               'location'               : self.project_repo                  ,
                               'buildspec'              : self.build_spec                   },
            'sourceVersion': self.source_version                                             ,
            'artifacts'    : {'type'                    : 'NO_ARTIFACTS'                    },
            'environment'  : {'type'                    : self.docker_type                   ,
                             'image'                    : f'{self.account_id}.dkr.ecr.{self.region}.amazonaws.com/{self.project_name}:latest'.lower(),
                             'computeType'              : self.compute_type                  ,
                             'imagePullCredentialsType' : 'SERVICE_ROLE'                    },
            'serviceRole'  : self.service_role
        }
        return self.code_build.codebuild.create_project(**kvargs)

    def policies__for_docker_build(self):
        cloud_watch_arn = f"arn:aws:logs:{self.region}:{self.account_id}:log-group:/aws/codebuild/{self.project_name}:log-stream:*"
        policies = {"Cloud-Watch-Policy": {"Version": "2012-10-17",
                                           "Statement": [{"Sid": "GsBotPolicy",
                                                          "Effect": "Allow",
                                                          "Action": ["logs:CreateLogGroup",
                                                                     "logs:CreateLogStream",
                                                                     "logs:PutLogEvents"],
                                                          "Resource": [ cloud_watch_arn ]}]},
                    "Secret-Manager": {
                        "Version": "2012-10-17",
                        "Statement": [{"Sid": "GsBotPolicy",
                                       "Effect": "Allow",
                                       "Action": ["secretsmanager:GetSecretValue", "secretsmanager:DescribeSecret"],
                                       "Resource": [
                                           f"arn:aws:secretsmanager:{self.region}:{self.account_id}:secret:slack-gs-bot-*",
                                           f"arn:aws:secretsmanager:{self.region}:{self.account_id}:secret:gw-elastic-server-*"]}]},  # todo: refactor
                    "Create-Docker-Image": {
                        "Version": "2012-10-17",
                        "Statement": [{"Effect": "Allow",
                                       "Action": ["ecr:BatchCheckLayerAvailability",
                                                  "ecr:CompleteLayerUpload",
                                                  "ecr:GetAuthorizationToken",
                                                  "ecr:InitiateLayerUpload",
                                                  "ecr:PutImage",
                                                  "ecr:UploadLayerPart"],
                                       "Resource": "*"}]},
                    "Create-Repository": {
                        "Version": "2012-10-17",
                        "Statement": [{"Effect": "Allow",
                                       "Action": ["ecr:CreateRepository"],
                                       "Resource": "*"}]}}
        return policies

    def policies__with_ecr(self):
        cloud_watch_arn = f"arn:aws:logs:{self.region}:{self.account_id}:log-group:/aws/codebuild/{self.project_name}:log-stream:*"
        policies = {"Cloud-Watch-Policy"     : { "Version": "2012-10-17",
                                                 "Statement": [{   "Effect": "Allow",
                                                                   "Action": [ "logs:CreateLogGroup"  ,
                                                                               "logs:CreateLogStream" ,
                                                                               "logs:PutLogEvents"   ],
                                                                   "Resource": [ cloud_watch_arn ]}]},
                    "Download_Image"         : { "Version": "2012-10-17",
                                                 "Statement": [{   "Effect": "Allow",
                                                                   "Action": [   "ecr:GetAuthorizationToken",
                                                                                 "ecr:BatchCheckLayerAvailability",
                                                                                 "ecr:GetDownloadUrlForLayer",
                                                                                 "ecr:GetRepositoryPolicy",
                                                                                 "ecr:DescribeRepositories",
                                                                                 "ecr:ListImages",
                                                                                 "ecr:DescribeImages",
                                                                                 "ecr:BatchGetImage"],
                                                                   "Resource": "*"}]}}
        return policies

    def policies__with_ecr_and_3_secrets(self):
        cloud_watch_arn = f"arn:aws:logs:{self.region}:{self.account_id}:log-group:/aws/codebuild/{self.project_name}:log-stream:*"
        policies = {"Cloud-Watch-Policy"     : { "Version": "2012-10-17",
                                                 "Statement": [{   "Sid": "GsBotPolicy",
                                                                   "Effect": "Allow",
                                                                   "Action": [ "logs:CreateLogGroup"  ,
                                                                               "logs:CreateLogStream" ,
                                                                               "logs:PutLogEvents"   ],
                                                                   "Resource": [ cloud_watch_arn ]}]},
                    "Download_Image"         : { "Version": "2012-10-17",
                                                 "Statement": [{   "Effect": "Allow",
                                                                   "Action": [   "ecr:GetAuthorizationToken",
                                                                                 "ecr:BatchCheckLayerAvailability",
                                                                                 "ecr:GetDownloadUrlForLayer",
                                                                                 "ecr:GetRepositoryPolicy",
                                                                                 "ecr:DescribeRepositories",
                                                                                 "ecr:ListImages",
                                                                                 "ecr:DescribeImages",
                                                                                 "ecr:BatchGetImage"],
                                                                   "Resource": "*"}]},
                    "Secret-Manager"        : { "Version"  : "2012-10-17",
                                                "Statement": [{   "Effect"  : "Allow",
                                                                  "Action"  : [  "secretsmanager:GetSecretValue","secretsmanager:DescribeSecret"          ],
                                                                  "Resource": [ f"arn:aws:secretsmanager:{self.region}:{self.account_id}:secret:slack-gs-bot-*"     ,
                                                                                f"arn:aws:secretsmanager:{self.region}:{self.account_id}:secret:gw-elastic*"        ,
                                                                                f"arn:aws:secretsmanager:{self.region}:{self.account_id}:secret:gsuite*"            ,
                                                                                f"arn:aws:secretsmanager:{self.region}:{self.account_id}:secret:git__*"         ]}]}}
        return policies


    def delete_project_role_and_policies(self):
        if self.code_build.iam.role_exists():
            policies_arns = list(self.code_build.iam.role_policies().values())
            self.code_build.iam.role_policies_detach(policies_arns)                 # first detach policies
            self.code_build.iam.policies_delete(policies_arns)                      # then delete the policies
            self.code_build.iam.role_delete()                                       # then delete the role
        if self.code_build.project_exists():
            self.code_build.project_delete()                                        # finally delete the project


import functools
from typing import get_type_hints, Union, Any


def enforce_type_safety(func):
    @functools.wraps(func)                                                                         # Preserve function's metadata during wrapping
    def wrapper(*args, **kwargs):                                                                  # Define the wrapper function to add type safety

        hints    = get_type_hints(func)                                                            # Retrieve type annotations of the function
        all_args = kwargs.copy()                                                                   # Start with keyword arguments
        all_args.update(dict(zip(func.__code__.co_varnames, args)))                                # Add positional arguments to the same dictionary

        for var_name, var_type in all_args.items():                                                # Iterate over all arguments provided to the function
            if var_name in hints:                                                                  # Check if the argument has a type hint specified
                expected_type = hints[var_name]                                                    # Get the expected type from type hints
                if expected_type is Any or expected_type is any:                                   # Skip type checking for Any, as it accepts any type
                    continue
                if hasattr(expected_type, '__origin__') and expected_type.__origin__ is Union:     # Check if the type hint is a Union
                    expected_types = expected_type.__args__                                        # Retrieve all types inside the Union
                else:
                    expected_types = (expected_type,)                                              # Wrap the expected type in a tuple for consistency

                if not isinstance(var_type, expected_types):                                       # Validate the type of the current argument
                    type_names    = ', '.join([t.__name__ for t in expected_types])                # Format the names of the expected types for the error message
                    error_message = (f"in '{func.__name__}' the '{var_name}' parameter "
                                     f"must be an '{type_names}', "
                                     f"but it was an '{type(var_type).__name__}'")                 # Construct the error message
                    raise TypeError(error_message)                                                 # Raise TypeError if the argument type does not match

        return func(*args, **kwargs)                                                               # Call the original function with the validated arguments
    return wrapper                                                                                 # Return the wrapped function



from functools import cache

from dotenv import load_dotenv

from osbot_aws.apis.Lambda import Lambda
from osbot_aws.aws.apigateway.API_Gateway_V2 import API_Gateway_V2
from osbot_aws.aws.iam.IAM_Assume_Role import IAM_Assume_Role

# todo add this permission to this role (required to create a lambda function
# {
#     "Version": "2012-10-17",
#     "Statement": [
#         {
#             "Sid": "VisualEditor0",
#             "Effect": "Allow",
#             "Action": "iam:PassRole",
#             "Resource": "*"
#         }
#     ]
# }
# todo add (ideally a more restricted version of)
# {
# 	"Version": "2012-10-17",
# 	"Statement": [
# 		{
# 			"Sid": "VisualEditor0",
# 			"Effect": "Allow",
# 			"Action": "s3:GetObject",
# 			"Resource": "*"
# 		}
# 	]
# }

class Lambda__with_temp_role(Lambda):
    temp_service_name   : str = "lambda"
    temp_role__services : str = ["lambda", 's3']
    temp_role__action   : str = "*"
    temp_role__resource : str = "*"
    temp_role__name     : str = 'osbot__temp_role_for__test_Lambda'
    temp_role__recreate : bool = False

    
    @cache
    def client(self):
        load_dotenv()
        service       = self.temp_service_name
        action        = self.temp_role__action
        resource      = self.temp_role__resource
        role_name     = self.temp_role__name
        role_recreate = self.temp_role__recreate
        policies_to_add = []
        for temp_role__service in self.temp_role__services:
            policies_to_add.append(dict(service=temp_role__service, action=action, resource=resource))

        iam_assume_role = IAM_Assume_Role(role_name=role_name, policies_to_add=policies_to_add)
        iam_assume_role.create_role(recreate=role_recreate)
        #iam_assume_role.credentials_reset()
        return iam_assume_role.boto3_client(service_name=service)

    def temp_role__iam_reset_credentials(self):
        iam_assume_role = IAM_Assume_Role(role_name=self.temp_role__name)
        return iam_assume_role.credentials_reset()


from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self

class Mistral_AI(Kwargs_To_Self):
    model_id          : str
    prompt            : str
    temperature       : float    = 0.0
    top_p             : float    = 1.0
    top_k             : int      = 1
    max_tokens        : int      = 1024
    stop              : list




    def body(self):
        return dict(prompt        = self.prompt         ,
                    temperature   = self.temperature    ,
                    top_p         = self.top_p          ,
                    top_k         = self.top_k          ,
                    max_tokens    = self.max_tokens     ,
                    stop          = self.stop           )



from osbot_aws.aws.dynamo_db.models.DyDB__Resource import DyDB__Resource


class Resource__Config(DyDB__Resource):

    def version(self):
        return self.document.get('version')

    def version__update(self, value):
        self.set_field('version', value)
        self.document['version'] = value


from botocore.client                                                    import BaseClient
from botocore.exceptions                                                import ClientError
from osbot_utils.helpers.sqlite.cache.Sqlite__Cache__Requests__Patch    import Sqlite__Cache__Requests__Patch
from osbot_utils.utils.Call_Stack import call_stack_frames_data
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Json                                             import json_to_str

SQLITE_DB_NAME__BOTO3_REQUESTS_CACHE = 'boto3_requests_cache.sqlite'
SQLITE_TABLE_NAME__BOTO3_REQUESTS    = 'boto3_requests'

class Cache_Boto3_Requests(Sqlite__Cache__Requests__Patch):

    def __init__(self, db_path=None):
        self.target_class                      = BaseClient
        self.target_function                   = BaseClient._make_api_call
        self.target_function_name              = "_make_api_call"
        self.db_name                           = SQLITE_DB_NAME__BOTO3_REQUESTS_CACHE
        self.table_name                        = SQLITE_TABLE_NAME__BOTO3_REQUESTS
        super().__init__(db_path=db_path)
        self.config.capture_exceptions         = True
        self.config.exception_classes          = [ClientError]
        self.add_caller_signature_to_cache_key = True                   # todo: move this to the main OSBot_Utils code base
        #self.print_requests         = False


    def cache_entry_comments(self, model_id, body):
        return super().cache_entry_comments(model_id=model_id, body=body)

    def cache_entry_comments_update(self, model_id, body, new_comments):
        return super().cache_entry_comments_update(new_comments, model_id=model_id, body=body)

    def invoke_target(self, target, target_args, target_kwargs):
        #if self.print_requests:
            #print(f'[invoke_target]: {target_args}')
        return super().invoke_target(target, target_args, target_kwargs)

    def request_data(self, *args, **kwargs):
        target_self, operation_name, api_params = args
        request_data =  {'operation_name': operation_name,
                        'api_params'    : api_params    }

        # need this in cases where we have multiple requests to the same method in the same tests
        if self.add_caller_signature_to_cache_key:                                     # todo: move this to the main OSBot_Utils code base
            frames_depth   = 15                                                        # todo find a better way to get the code that is changing
            frames_capture = frames_depth
            caller_signature = ''
            frames              = call_stack_frames_data(frames_depth)                 # todo: refactor this to separate method
            for frame in frames[0:frames_capture]:
                caller_signature += f"{frame.get('name')}:{frame.get('lineno')} | "
            request_data['caller_signature'] = caller_signature                      # this adds support for different caches to the same method call (main limitation is that it is directly tied with the line numbers)

        request_data = json_to_str(request_data)            # todo: this used to use yaml, change to a better mode
        return request_data


from typing                             import Callable
from osbot_aws.apis.Lambda              import Lambda
from osbot_aws.deploy.Deploy_Lambda     import Deploy_Lambda
from osbot_aws.helpers.Lambda_Upload_Package import Lambda_Upload_Package
from osbot_aws.testing.TestCase__Lambda import TestCase__Lambda


class TestCase__Lambda__Deploy(TestCase__Lambda):
    aws_lambda    : Lambda
    handler       : Callable
    deploy_lambda : Deploy_Lambda

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.deploy_lambda    = Deploy_Lambda(cls.handler)
        cls.aws_lambda       = cls.deploy_lambda.lambda_function()
        cls.aws_lambda.client = cls.client__lambda                          # need to replace this so that it uses the classes that have
        cls.aws_lambda.s3     = cls.type_S3                                 # the temp credentials

    def deploy(self):
        return self.deploy_lambda.update()

    def invoke(self, payload):
        return self.deploy_lambda.invoke(payload)

    def invoke_return_logs(self, payload):
        return self.deploy_lambda.invoke_return_logs(payload)

    def upload_dependencies_to_s3(self, dependencies):
        lambda_upload_package = Lambda_Upload_Package()
        lambda_upload_package.dependencies.s3 = self.type_S3
        lambda_upload_package.upload_to_s3(dependencies)


from osbot_utils.utils.Misc import wait_for

from osbot_utils.decorators.methods.cache_on_self import cache_on_self

from osbot_utils.utils.Dev import pprint

from osbot_utils.decorators.methods.remove_return_value import remove_return_value

from osbot_utils.decorators.lists.group_by import group_by
from osbot_utils.decorators.lists.index_by import index_by
from osbot_utils.decorators.methods.catch import catch

from osbot_aws.apis.Session            import Session


class Route_53:

    @cache_on_self
    def client(self):
        return Session().client('route53',region_name='us-east-1')

    @cache_on_self
    def client_domains(self):
        return Session().client('route53domains',region_name='us-east-1')

    # Main methods

    def change_details(self, change_id):
        response = self.client().get_change(Id=change_id)
        return response.get('ChangeInfo')

    @catch
    @index_by
    def domains(self):
        return self.client_domains().list_domains().get('Domains')

    @remove_return_value('ResponseMetadata')
    def hosted_zone(self, hosted_zone_id):
        return self.client().get_hosted_zone(Id= hosted_zone_id)

    @catch
    @index_by
    def hosted_zones(self):
        return self.client().list_hosted_zones().get('HostedZones')

    def record_set__change(self, hosted_zone_id, action, resource_record_set):
        change = { "Action"           : action,
                   "ResourceRecordSet": resource_record_set }
        return self.record_set__changes(hosted_zone_id, [change])

    def record_set__changes(self, hosted_zone_id, changes):
        kwargs = {  "HostedZoneId" : hosted_zone_id,
                    "ChangeBatch": { "Changes": changes }}

        return self.client().change_resource_record_sets(**kwargs)

    def record_set__delete__alias(self, hosted_zone_id, record_name, record_type, alias_dns_name, alias_hosted_zone_id):
        resource_record_set = dict(Name        = record_name,
                                   Type        = record_type,
                                   AliasTarget = dict(DNSName              = alias_dns_name      ,
                                                      HostedZoneId         = alias_hosted_zone_id,
                                                      EvaluateTargetHealth = False               ))
        kwargs = dict(hosted_zone_id      = hosted_zone_id      ,
                      action              = 'DELETE'            ,
                      resource_record_set = resource_record_set )

        return self.record_set__change(**kwargs)

    @catch
    def record_set__upsert(self, record_name, record_type, alias_target,hosted_zone_id,alias_hosted_zone_id):
        params = {  "HostedZoneId" : hosted_zone_id,
                    "ChangeBatch": { "Changes": [{ "Action": "UPSERT",
                                                   "ResourceRecordSet": { "Name": record_name,
                                                                          "Type": record_type,
                                                                          "AliasTarget": { "DNSName"             : alias_target         ,
                                                                                           "HostedZoneId"        : alias_hosted_zone_id ,
                                                                                           "EvaluateTargetHealth": False                }}}]}}
        return self.client().change_resource_record_sets(**params)



    def record_set(self, hosted_zone_id, record_name, record_type):
        kwargs = dict(HostedZoneId    = hosted_zone_id,
                      StartRecordName = record_name   ,                                         # this is a trick to get the search
                      StartRecordType = record_type   ,                                         # since the search is 'starts with'
                      MaxItems        = "1"           )                                         # we ask for just one


        record_sets = self.client().list_resource_record_sets(**kwargs).get('ResourceRecordSets')

        if len(record_sets) == 1:                                                               # if we got an answer
            record_set = record_sets[0]                                                         # get the first
            record_set_name = record_set.get('Name')                                            # get the record_set name
            if record_set_name == record_name or record_set_name == record_name + '.':          # if the name matches the provided name or the provided name + '.'
                return    record_set                                                            # return it
        return {}

    @catch
    @group_by
    def record_sets(self, hosted_zone_id):
        return self.client().list_resource_record_sets(HostedZoneId=hosted_zone_id).get('ResourceRecordSets')

    @remove_return_value('ResponseMetadata')
    def check_dns_answer(self, hosted_zone_id, record_name, record_type='A'):
        kwargs = dict(HostedZoneId = hosted_zone_id ,
                      RecordName   = record_name    ,
                      RecordType   = record_type    ) #'SOA' | 'A' | 'TXT' | 'NS' | 'CNAME' | 'MX' | 'NAPTR' | 'PTR' | 'SRV' | 'SPF' | 'AAAA' | 'CAA' | 'DS',)
        return self.client().test_dns_answer(**kwargs)

    # this can actulaly take quite a long time, some DNS changes take more than 1 minute
    def wait_for_change_status(self, change_id, status, max_attempts=30, sleep_for=1):
        for i in range(0, max_attempts):
            response       = self.change_details(change_id)
            current_status = response.get('Status')
            if current_status == status:
                return True
            wait_for(sleep_for)
        return False

    def wait_for_change_completed(self, change_id, max_attempts=30):
        return self.wait_for_change_status(change_id, 'INSYNC', max_attempts=max_attempts)


from osbot_utils.base_classes.Type_Safe import Type_Safe
from osbot_utils.utils.Misc import random_text, lower

TEMP_ROLE__NAME_FORMAT = 'ROLE__temp_osbot__{role_access_type}__{boto3_service_name}'

class Temp_Role__For_Service__Config(Type_Safe):
    boto3_service_name : str
    required_services  : list
    action             : str = "*"
    resource           : str = "*"
    recreate           : bool = False
    random_role_name   : bool = False
    role_name          : str
    role_access_type   : str = 'Full_Access'

    def data(self):
        if not self.boto3_service_name:
            raise ValueError("boto3_service_name is required")
        if not self.required_services:
            raise ValueError("required_services is required")
        if not self.role_name:
            self.role_name = TEMP_ROLE__NAME_FORMAT.format(role_access_type=self.role_access_type, boto3_service_name=self.boto3_service_name)
            if self.random_role_name is True:
                self.role_name += lower(random_text('_'))
        return self


import re


def str_to_valid_s3_bucket_name(value):

    s3_bucket_name = re.sub(r'[^a-z0-9-]', '', value.lower()).strip('.-')     # Replace invalid characters with a hyphen and ensure the name does not start or end with invalid characters

    if s3_bucket_name.endswith('-'):
        s3_bucket_name = s3_bucket_name[:-1]                    # Remove trailing hyphen if present
    if not s3_bucket_name:
        raise ValueError(f"[osbot_aws.utils.AWS_Sanitization] in str_to_valid_s3_bucket_name, could not create a valid name from: {value}")
    return s3_bucket_name



from osbot_aws.aws.iam.roles.IAM__Statement__Resource import IAM__Statement__Resource


class IAM__Statement__Resource__Allow(IAM__Statement__Resource):
    effect   : str = 'Allow'

    def __init__(self, actions, resource):
        self.actions  = actions
        self.resource = resource
        super().__init__()